#!/usr/bin/python
#dhnnet.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#    some network routines
# 

import os
import sys
import subprocess
import types
import socket
import urllib
import urllib2
import urlparse
import mimetypes

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in dhnnet.py')

from twisted.internet.defer import Deferred, DeferredList, succeed, fail
from twisted.internet import ssl
from twisted.internet import protocol
# from twisted.internet.utils import getProcessOutput
from twisted.protocols import basic
from twisted.web import iweb
from twisted.web import client
from twisted.web import http_headers
from twisted.web import http
from twisted.web.client import getPage
from twisted.web.client import downloadPage
from twisted.web.client import HTTPClientFactory
from twisted.web.client import HTTPDownloader

from zope.interface import implements

from OpenSSL import SSL


_ConnectionDoneCallbackFunc = None
_ConnectionFailedCallbackFunc = None

#------------------------------------------------------------------------------ 

_UserAgentString = "DataHaven.NET-http-agent"
_ProxySettings = {
    'host': '',
    'port': '',
    'ssl': 'False',
    'username':'',
    'password': ''
    }

#------------------------------------------------------------------------------

def init():
    pass

def SetConnectionDoneCallbackFunc(f):
    global _ConnectionDoneCallbackFunc
    _ConnectionDoneCallbackFunc = f
    
def SetConnectionFailedCallbackFunc(f):
    global _ConnectionFailedCallbackFunc
    _ConnectionFailedCallbackFunc = f

def ConnectionDone(param=None, proto=None, info=None):
    global _ConnectionDoneCallbackFunc
    if _ConnectionDoneCallbackFunc is not None:
        _ConnectionDoneCallbackFunc(proto, info, param)
    return param

def ConnectionFailed(param=None, proto=None, info=None):
    global _ConnectionFailedCallbackFunc
    if _ConnectionFailedCallbackFunc is not None:
        _ConnectionFailedCallbackFunc(proto, info, param)
    return param

#------------------------------------------------------------------------------ 

def parseurl(url, defaultPort=None):
    """Split the given URL into the scheme, host, port, and path."""
    url = url.strip()
    parsed = urlparse.urlparse(url)
    scheme = parsed[0]
    path = urlparse.urlunparse(('', '') + parsed[2:])
    if defaultPort is None:
        if scheme == 'https':
            defaultPort = 443
        else:
            defaultPort = 80
    host, port = parsed[1], defaultPort
    if ':' in host:
        host, port = host.split(':')
        try:
            port = int(port)
        except ValueError:
            port = defaultPort
    if path == '':
        path = '/'
    return scheme, host, port, path

#------------------------------------------------------------------------------ 

def detect_proxy_settings():
    d = {
        'host':'',
        'port':'',
        'username':'',
        'password':'',
        'ssl': 'False'}
    httpproxy = urllib2.getproxies().get('http', None)
    httpsproxy = urllib2.getproxies().get('https', None)

    if httpproxy is not None:
        try:
            scheme, host, port, path = parseurl(httpproxy)
        except:
            return d
        d['ssl'] = 'False'
        d['host'] = host
        d['port'] = port

    if httpsproxy is not None:
        try:
            scheme, host, port, path = parseurl(httpsproxy)
        except:
            return d
        d['ssl'] = 'True'
        d['host'] = host
        d['port'] = port

    return d

def set_proxy_settings(settings_dict):
    global _ProxySettings
    _ProxySettings = settings_dict

def get_proxy_settings():
    global _ProxySettings
    return _ProxySettings

def get_proxy_host():
    global _ProxySettings
    return _ProxySettings.get('host', '')

def get_proxy_port():
    global _ProxySettings
    try:
        return int(_ProxySettings.get('port', '8080'))
    except:
        return 8080

def get_proxy_username():
    global _ProxySettings
    return _ProxySettings.get('username', '')

def get_proxy_password():
    global _ProxySettings
    return _ProxySettings.get('password', '')

def get_proxy_ssl():
    global _ProxySettings
    return _ProxySettings.get('ssl', '')

def proxy_is_on():
    return get_proxy_host() != ''

#-------------------------------------------------------------------------------

class MyHTTPClientFactory(HTTPClientFactory):
    def page(self, page):
        if self.waiting:
            self.waiting = 0
            self.deferred.callback(page)

    def noPage(self, reason):
        if self.waiting:
            self.waiting = 0
            self.deferred.errback(reason)

    def clientConnectionFailed(self, _, reason):
        if self.waiting:
            self.waiting = 0
            self.deferred.errback(reason)


def getPageTwistedOld(url):
    global _UserAgentString
    scheme, host, port, path = parseurl(url)
    factory = MyHTTPClientFactory(url, agent=_UserAgentString)
    reactor.connectTCP(host, port, factory)
    return factory.deferred

#------------------------------------------------------------------------------


def downloadPageTwisted(url, filename):
    global _UserAgentString
    return downloadPage(url, filename, agent=_UserAgentString)

#-------------------------------------------------------------------------------

#http://schwerkraft.elitedvb.net/plugins/scmcvs/cvsweb.php/enigma2-plugins/mediadownloader/src/HTTPProgressDownloader.py?rev=1.1;cvsroot=enigma2-plugins;only_with_tag=HEAD
class HTTPProgressDownloader(HTTPDownloader):
    """Download to a file and keep track of progress."""

    def __init__(self, url, fileOrName, writeProgress = None, *args, **kwargs):
        HTTPDownloader.__init__(self, url, fileOrName, supportPartial=0, *args, **kwargs)
        # Save callback(s) locally
        if writeProgress and type(writeProgress) is not list:
            writeProgress = [ writeProgress ]
        self.writeProgress = writeProgress

        # Initialize
        self.currentlength = 0
        self.totallength = None

    def gotHeaders(self, headers):
        HTTPDownloader.gotHeaders(self, headers)

        # If we have a callback and 'OK' from Server try to get length
        if self.writeProgress and self.status == '200':
            if headers.has_key('content-length'):
                self.totallength = int(headers['content-length'][0])
                for cb in self.writeProgress:
                    if cb:
                        cb(0, self.totallength)


    def pagePart(self, data):
        HTTPDownloader.pagePart(self, data)

        # If we have a callback and 'OK' from server increment pos
        if self.writeProgress and self.status == '200':
            self.currentlength += len(data)
            for cb in self.writeProgress:
                if cb:
                    cb(self.currentlength, self.totallength)




def downloadWithProgressTwisted(url, file, progress_func):
    global _UserAgentString
    scheme, host, port, path = parseurl(url)
    factory = HTTPProgressDownloader(url, file, progress_func, agent=_UserAgentString)
    if scheme == 'https':
        contextFactory = ssl.ClientContextFactory()
        reactor.connectSSL(host, port, factory, contextFactory)
    else:
        reactor.connectTCP(host, port, factory)
    return factory.deferred

#-------------------------------------------------------------------------------


def downloadSSLWithProgressTwisted(url, file, progress_func, privateKeyFileName, certificateFileName):
    global _UserAgentString
    scheme, host, port, path = parseurl(url)
    factory = HTTPProgressDownloader(url, file, progress_func, agent=_UserAgentString)
    if scheme != 'https':
        return None
    contextFactory = ssl.DefaultOpenSSLContextFactory(privateKeyFileName, certificateFileName)
    reactor.connectSSL(host, port, factory, contextFactory)
    return factory.deferred


#-------------------------------------------------------------------------------


class MyClientContextFactory(ssl.ClientContextFactory):
    def __init__(self, certificates_filenames):
        self.certificates_filenames = list(certificates_filenames)

    def verify(self, connection, x509, errnum, errdepth, ok):
        return ok

    def getContext(self):
        ctx = ssl.ClientContextFactory.getContext(self)
        for cert in self.certificates_filenames:
            try:
                ctx.load_verify_locations(cert)
            except:
                pass
        ctx.set_verify(SSL.VERIFY_PEER | SSL.VERIFY_FAIL_IF_NO_PEER_CERT , self.verify)
        return ctx


def downloadSSL(url, fileOrName, progress_func, certificates_filenames):
    global _UserAgentString
    scheme, host, port, path = parseurl(url)
    if not isinstance(certificates_filenames, types.ListType):
        certificates_filenames = [certificates_filenames, ]
    cert_found = False
    for cert in certificates_filenames:
        if os.path.isfile(cert) and os.access(cert, os.R_OK):
            cert_found = True
            break
    if not cert_found:
        return fail('no one ssl certificate found')
    factory = HTTPDownloader(url, fileOrName, agent=_UserAgentString)
    contextFactory = MyClientContextFactory(certificates_filenames)
    reactor.connectSSL(host, port, factory, contextFactory)
    return factory.deferred

#------------------------------------------------------------------------------

class NoVerifyClientContextFactory:
    isClient = 1
    method = SSL.SSLv3_METHOD
    def getContext(self):
        def x(*args):
            return True
        ctx = SSL.Context(self.method)
        #print dir(ctx)
        ctx.set_verify(SSL.VERIFY_NONE,x)
        return ctx

def downloadSSL2(url, fileOrName, progress_func, certificates_filenames=[]):
    global _UserAgentString
    factory = HTTPDownloader(url, fileOrName, agent=_UserAgentString)

    if proxy_is_on():
        factory.path = url
        # TODO
        # need to check certificate too
        contextFactory = MyClientContextFactory(certificates_filenames)
        reactor.connectSSL(get_proxy_host(), get_proxy_port(), factory, contextFactory)

    else:
        scheme, host, port, path = parseurl(url)
        if not isinstance(certificates_filenames, types.ListType):
            certificates_filenames = [certificates_filenames, ]
        cert_found = False
        for cert in certificates_filenames:
            if os.path.isfile(cert) and os.access(cert, os.R_OK):
                cert_found = True
                break
        if not cert_found:
            return fail('no one ssl certificate found')
        contextFactory = MyClientContextFactory(certificates_filenames)
        reactor.connectSSL(host, port, factory, contextFactory)

    return factory.deferred

#-------------------------------------------------------------------------------

class ProxyClientFactory(client.HTTPClientFactory):
    def setURL(self, url):
        client.HTTPClientFactory.setURL(self, url)
        self.path = url

def getPageTwisted(url, timeout=0):
    global _UserAgentString
    if proxy_is_on():
        factory = ProxyClientFactory(url, agent=_UserAgentString, timeout=timeout)
        reactor.connectTCP(get_proxy_host(), get_proxy_port(), factory)
        factory.deferred.addCallback(ConnectionDone, 'http', 'getPageTwisted proxy %s' % (url))
        factory.deferred.addErrback(ConnectionFailed, 'http', 'getPageTwisted proxy %s' % (url))
        return factory.deferred
    else:
        d = getPage(url, agent=_UserAgentString, timeout=timeout)
        d.addCallback(ConnectionDone, 'http', 'getPageTwisted %s' % url)
        d.addErrback(ConnectionFailed, 'http', 'getPageTwisted %s' % url)
        return d

#------------------------------------------------------------------------------

def downloadHTTP(url, fileOrName):
    global _UserAgentString
    scheme, host, port, path = parseurl(url)
    factory = HTTPDownloader(url, fileOrName, agent=_UserAgentString)
    if proxy_is_on():
        host = get_proxy_host()
        port = get_proxy_port()
        factory.path = url
    reactor.connectTCP(host, port, factory)
    return factory.deferred

#-------------------------------------------------------------------------------

def IpIsLocal(ip):
    if ip == '':
        return True
    if ip == '0.0.0.0':
        return True
    if ip.startswith('192.168.'):
        return True
    if ip.startswith('10.'):
        return True
    if ip.startswith('127.'):
        return True
    if ip.startswith('172.'):
        try:
            secondByte = int(ip.split('.')[1])
        except:
            raise Exception('wrong ip address ' + str(ip))
            return True
        if secondByte >= 16 and secondByte <= 31:
            return True
    return False

#-------------------------------------------------------------------------------

def getLocalIpError():
    import socket
    addr = socket.gethostbyname(socket.gethostname())
    if addr == "127.0.0.1" and os.name == 'posix':
        import commands
        output = commands.getoutput("/sbin/ifconfig")
        #TODO parseaddress not done yet
        addr = parseaddress(output)
    return addr


#http://ubuntuforums.org/showthread.php?t=1215042
def getLocalIp(): # had this in p2p/stun.py
    import socket
    # 1: Use the gethostname method

    try:
        ipaddr = socket.gethostbyname(socket.gethostname())
        if not( ipaddr.startswith('127') ) :
            #print('Can use Method 1: ' + ipaddr)
            return ipaddr
    except:
        pass

    # 2: Use outside connection
    '''
    Source:
    http://commandline.org.uk/python/how-to-find-out-ip-address-in-python/
    '''

    ipaddr=''
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('google.com', 0))
        ipaddr=s.getsockname()[0]
        #print('Can used Method 2: ' + ipaddr)
        return ipaddr
    except:
        pass


    # 3: Use OS specific command
    import subprocess , platform
    ipaddr=''
    os_str=platform.system().upper()

    if os_str=='LINUX' :

        # Linux:
        arg='ip route list'
        p=subprocess.Popen(arg,shell=True,stdout=subprocess.PIPE)
        data = p.communicate()
        sdata = data[0].split()
        ipaddr = sdata[ sdata.index('src')+1 ]
        #netdev = sdata[ sdata.index('dev')+1 ]
        #print('Can used Method 3: ' + ipaddr)
        return ipaddr

    elif os_str=='WINDOWS' :

        # Windows:
        arg='route print 0.0.0.0'
        p=subprocess.Popen(arg,shell=True,stdout=subprocess.PIPE)
        data = p.communicate()
        strdata=data[0].decode()
        sdata = strdata.split()

        while len(sdata)>0:
            if sdata.pop(0)=='Netmask' :
                if sdata[0]=='Gateway' and sdata[1]=='Interface' :
                    ipaddr=sdata[6]
                    break
        #print('Can used Method 4: ' + ipaddr)
        return ipaddr

    return '127.0.0.1' # uh oh, we're in trouble, but don't want to return none

#-------------------------------------------------------------------------------

def TestInternetConnectionOld(remote_host='www.google.com'): # 74.125.113.99
#    return True

    try:
        (family, socktype, proto, garbage, address) = socket.getaddrinfo(remote_host, "http")[0]
    except Exception, e:
        return False

    s = socket.socket(family, socktype, proto)

    try:
        result = s.connect(address)
    except Exception, e:
        return False

    return result is None or result == 0

#------------------------------------------------------------------------------ 

def TestInternetConnectionOld2(remote_hosts=None, timeout=10):
    if remote_hosts is None:
        remote_hosts = []
        remote_hosts.append('www.google.com')
        remote_hosts.append('www.facebook.com')
        remote_hosts.append('www.youtube.com')
        # remote_hosts.append('www.yahoo.com')
        # remote_hosts.append('www.baidu.com')
    def _response(src, result):
        print '_response', err, hosts, index
        result.callback(None)
    def _fail(err, hosts, index, result):
        print 'fail', hosts, index
        reactor.callLater(0, _call, hosts, index+1, result)
    def _call(hosts, index, result):
        print 'call' , hosts, index, result
        if index >= len(hosts):
            result.errback(None)
            return
        print '    ', hosts[index]
        d = getPageTwisted(hosts[index])
        d.addCallback(_response, result)
        d.addErrback(_fail, hosts, index, result)
    result = Deferred()
    reactor.callLater(0, _call, remote_hosts, 0, result)
    return result
        
#------------------------------------------------------------------------------  

def TestInternetConnection(remote_hosts=None, timeout=10):
    if remote_hosts is None:
        remote_hosts = []
        remote_hosts.append('http://www.google.com')
        remote_hosts.append('http://www.facebook.com')
        remote_hosts.append('http://www.youtube.com')
        # remote_hosts.append('www.yahoo.com')
        # remote_hosts.append('www.baidu.com')
    dl = []
    for host in remote_hosts:
        dl.append(getPageTwisted(host, timeout))
    return DeferredList(dl, fireOnOneCallback=True, fireOnOneErrback=False, consumeErrors=True)
        
#------------------------------------------------------------------------------ 

def SendEmail(TO, FROM, HOST, PORT, LOGIN, PASSWORD, SUBJECT, BODY, FILES):
#    try:
        import smtplib
        from email import Encoders
        from email.MIMEText import MIMEText
        from email.MIMEBase import MIMEBase
        from email.MIMEMultipart import MIMEMultipart
        from email.Utils import formatdate

        msg = MIMEMultipart()
        msg["From"] = FROM
        msg["To"] = TO
        msg["Subject"] = SUBJECT
        msg["Date"]    = formatdate(localtime=True)
        msg.attach(MIMEText(BODY))

        # attach a file
        for filePath in FILES:
            if not os.path.isfile(filePath):
                continue
            part = MIMEBase('application', "octet-stream")
            part.set_payload( open(filePath,"rb").read() )
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(filePath))
            msg.attach(part)

        s = smtplib.SMTP(HOST, PORT)

        #s.set_debuglevel(True) # It's nice to see what's going on

        s.ehlo() # identify ourselves, prompting server for supported features

        if s.has_extn('STARTTLS'):
            s.starttls()
            s.ehlo() # re-identify ourse

        s.login(LOGIN, PASSWORD)  # optional

        failed = s.sendmail(FROM, TO, msg.as_string())

        s.close()

#    except:
#        dhnio.DprintException()


#-------------------------------------------------------------------------------

# Great Thanks to Mariano!
# http://marianoiglesias.com.ar/python/file-uploading-with-multi-part-encoding-using-twisted/

class StringReceiver(protocol.Protocol):
    buffer = ""

    def __init__(self, deferred=None):
        self._deferred = deferred

    def dataReceived(self, data):
        self.buffer += data

    def connectionLost(self, reason):
        if self._deferred and reason.check(client.ResponseDone):
            self._deferred.callback(self.buffer)
        else:
            self._deferred.errback(Exception(self.buffer))
            
class MultiPartProducer:
    implements(iweb.IBodyProducer)
    CHUNK_SIZE = 2 ** 8
    def __init__(self, files={}, data={}, callback=None, deferred=None):
        self._files = files
        self._file_lengths = {}
        self._data = data
        self._callback = callback
        self._deferred = deferred
        self.boundary = self._boundary()
        self.length = self._length()
        
    def startProducing(self, consumer):
        self._consumer = consumer
        self._current_deferred = Deferred()
        self._sent = 0
        self._paused = False
        if not hasattr(self, "_chunk_headers"):
            self._build_chunk_headers()
        if self._data:
            block = ""
            for field in self._data:
                block += self._chunk_headers[field]
                block += self._data[field]
                block += "\r\n"
            self._send_to_consumer(block)
        if self._files:
            self._files_iterator = self._files.iterkeys()
            self._files_sent = 0
            self._files_length = len(self._files)
            self._current_file_path = None
            self._current_file_handle = None
            self._current_file_length = None
            self._current_file_sent = 0
            result = self._produce()
            if result:
                return result
        else:
            return succeed(None)
        return self._current_deferred
    
    def resumeProducing(self):
        self._paused = False
        result = self._produce()
        if result:
            return result
        
    def pauseProducing(self):
        self._paused = True
        
    def stopProducing(self):
        self._finish(True)
        if self._deferred and self._sent < self.length:
            self._deferred.errback(Exception("Consumer asked to stop production of request body (%d sent out of %d)" % (self._sent, self.length)))
            
    def _produce(self):
        if self._paused:
            return
        done = False
        while not done and not self._paused:
            if not self._current_file_handle:
                field = self._files_iterator.next()
                self._current_file_path = self._files[field]
                self._current_file_sent = 0
                self._current_file_length = self._file_lengths[field]
                self._current_file_handle = open(self._current_file_path, "rb")
                self._send_to_consumer(self._chunk_headers[field])
            chunk = self._current_file_handle.read(self.CHUNK_SIZE)
            if chunk:
                self._send_to_consumer(chunk)
                self._current_file_sent += len(chunk)
            if not chunk or self._current_file_sent == self._current_file_length:
                self._send_to_consumer("\r\n")
                self._current_file_handle.close()
                self._current_file_handle = None
                self._current_file_sent = 0
                self._current_file_path = None
                self._files_sent += 1
            if self._files_sent == self._files_length:
                done = True
        if done:
            self._send_to_consumer("--%s--\r\n" % self.boundary)
            self._finish()
            return succeed(None)
        
    def _finish(self, forced=False):
        if hasattr(self, "_current_file_handle") and self._current_file_handle:
            self._current_file_handle.close()
        if self._current_deferred:
            self._current_deferred.callback(self._sent)
            self._current_deferred = None
        if not forced and self._deferred:
            self._deferred.callback(self._sent)
            
    def _send_to_consumer(self, block):
        self._consumer.write(block)
        self._sent += len(block)
        if self._callback:
            self._callback(self._sent, self.length)
            
    def _length(self):
        self._build_chunk_headers()
        length = 0
        if self._data:
            for field in self._data:
                length += len(self._chunk_headers[field])
                length += len(self._data[field])
                length += 2
        if self._files:
            for field in self._files:
                length += len(self._chunk_headers[field])
                length += self._file_size(field)
                length += 2
        length += len(self.boundary)
        length += 6
        return length
    
    def _build_chunk_headers(self):
        if hasattr(self, "_chunk_headers") and self._chunk_headers:
            return
        self._chunk_headers = {}
        for field in self._files:
            self._chunk_headers[field] = self._headers(field, True)
        for field in self._data:
            self._chunk_headers[field] = self._headers(field)
            
    def _headers(self, name, is_file=False):
        value = self._files[name] if is_file else self._data[name]
        _boundary = self.boundary.encode("utf-8") if isinstance(self.boundary, unicode) else urllib.quote_plus(self.boundary)
        headers = ["--%s" % _boundary]
        if is_file:
            disposition = 'form-data; name="%s"; filename="%s"' % (name, os.path.basename(value))
        else:
            disposition = 'form-data; name="%s"' % name
        headers.append("Content-Disposition: %s" % disposition)
        if is_file:
            file_type = self._file_type(name)
        else:
            file_type = "text/plain; charset=utf-8"
        headers.append("Content-Type: %s" % file_type)
        if is_file:
            headers.append("Content-Length: %i" % self._file_size(name))
        else:
            headers.append("Content-Length: %i" % len(value))
        headers.append("")
        headers.append("")
        return "\r\n".join(headers)
    
    def _boundary(self):
        boundary = None
        try:
            import uuid
            boundary = uuid.uuid4().hex
        except ImportError:
            import random, sha
            bits = random.getrandbits(160)
            boundary = sha.new(str(bits)).hexdigest()
        return boundary
    
    def _file_type(self, field):
        type = mimetypes.guess_type(self._files[field])[0]
        return type.encode("utf-8") if isinstance(type, unicode) else str(type)
    
    def _file_size(self, field):
        size = 0
        try:
            handle = open(self._files[field], "r")
            size = os.fstat(handle.fileno()).st_size
            handle.close()
        except:
            size = 0
        self._file_lengths[field] = size
        return self._file_lengths[field]
       
#------------------------------------------------------------------------------ 

def uploadHTTP(url, files, data, progress=None, receiverDeferred=None):
    # producerDeferred = Deferred()
    receiverDeferred = Deferred()
    
    myProducer = MultiPartProducer(files, data, progress) #, producerDeferred)
    myReceiver = StringReceiver(receiverDeferred)
    
    headers = http_headers.Headers()
    headers.addRawHeader("Content-Type", "multipart/form-data; boundary=%s" % myProducer.boundary)
    
    agent = client.Agent(reactor)
    request = agent.request("POST", url, headers, myProducer)
    request.addCallback(lambda response: response.deliverBody(myReceiver))
    
#------------------------------------------------------------------------------ 

# return a list of IPs for current active network interfaces 
def getNetworkInterfaces():
    import platform
    plat = platform.uname()[0]
    
    if plat == 'Windows':
        import re
        dirs = ['', r'c:\windows\system32', r'c:\winnt\system32']
        try:
            import ctypes
            buffer = ctypes.create_string_buffer(300)
            ctypes.windll.kernel32.GetSystemDirectoryA(buffer, 300)
            dirs.insert(0, buffer.value.decode('mbcs'))
        except:
            return []
        for dir in dirs:
            try:
                pipe = os.popen(os.path.join(dir, 'ipconfig') + ' /all')
            except IOError:
                return []
            rawtxt = unicode(pipe.read())
            ips_unicode = re.findall(u'^.*?IP.*?(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).*$', rawtxt, re.U | re.M)
            ips = []
            for ip in ips_unicode:
                ips.append(str(ip))
            del ips_unicode
            return ips
        
#    elif plat == 'Linux':
#        try:
#            pipe = os.popen("/bin/ip -f inet a")
#        except IOError:
#            return []
#        print 'getNetworkInterfaces', pipe
#        try:
#            rawtxt = unicode(pipe.read())
#            lines = rawtxt.splitlines()
#        except:
#            return []
#        print 'getNetworkInterfaces:\n', rawtext
#        ips = set()
#        # ifaces = {}
#        for line in lines:
#            check = line.strip('\n').strip().split(' ')
#            if check[0] == "inet":
#                if check[2] == "brd":
#                    check.pop(2)
#                    check.pop(2)
#                # if not ifaces.has_key(check[4]):
#                #     ifaces[check[4]] = []
#                ipaddress = check[1].split("/")[0]
#                # ifaces[check[4]].append(ipaddress)
#                ips.add(ipaddress)
#        print 'getNetworkInterfaces', ips
#        return list(ips)
    
    elif plat == 'Linux':
        import re
        ips = []
        output = subprocess.Popen(['ifconfig',], stdout=subprocess.PIPE).communicate()[0]
        # print 'getNetworkInterfaces:\n', output
        for interface in output.split('\n\n'):
            if interface.strip():
                ipaddr = re.search(r'inet addr:(\S+)', interface)
                if ipaddr:
                    ips.append(ipaddr.group(1))
        return ips
    
    else:
        return []

#------------------------------------------------------------------------------ 


def test1():
    def done(x, filen):
        print 'done'
        print x
        print open(filen).read()
    def fail(x):
        print 'fail'
        print x.getErrorMessage()

    url = 'https://downloads.datahaven.net/info.txt'
    fn = '123'
    d = downloadSSL1(url, fn, None, 'dhn.cer')
    d.addCallback(done, fn)
    d.addErrback(fail)

def test2():
    def done(x):
        #print 'done!!!!!!!!!!!!!!!!!!!!!!!!!!!'
        print x
    def fail(x):
        print 'fail'
        print x
        print x.getErrorMessage()
    url = 'http://identity.datahaven.net/veselin.xml'
    d = getPageTwisted(url)
    d.addCallback(done)
    d.addErrback(fail)

def test3():
    d = detect_proxy_settings()
    print d

def test4():
    def done(x, filen):
        print 'done'
        print x
        print open(filen).read()
    def fail(x):
        print 'fail'
        print x.getErrorMessage()

#    import settings
#    url = settings.UpdateLocationURL() + settings.CurrentVersionDigestsFilename()
    url = 'http://identity.datahaven.net/downloads/version.txt'
    fn = '123'
    d = downloadHTTP(url, fn)
    d.addCallback(done, fn)
    d.addErrback(fail)

def test5():
    SendEmail('to.datahaven@gmail.com',
              'dhnemail1@mail.offshore.ai',
              'smtp.gmail.com',
              587,
              'datahaven.net.mail@gmail.com',
              'datahaven.net.mail',
              'subj',
              'some body \n\n body',
              ['__init__.pyc'],)


def test6():
    import dhnio
    dhnio.init()
    print getNetworkInterfaces()
    
    
def test7():
    import dhnio
    dhnio.init()
    def _ok(x):
        print x
        reactor.stop()
    def _fail(x):
        print x
        reactor.stop()
    TestInternetConnection().addCallbacks(_ok, _fail)
    reactor.run()

#------------------------------------------------------------------------------ 

if __name__ == '__main__':
    #init()
    test7()
    #reactor.run()





import os
import binascii
#import warnings
#from Crypto.pct_warnings import PowmInsecureWarning
#warnings.filterwarnings("ignore", category=PowmInsecureWarning)
from Crypto.PublicKey import RSA
from asn1 import DerSequence
from OpenSSL.crypto import load_privatekey, dump_privatekey, FILETYPE_PEM, FILETYPE_ASN1
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.der import decoder as der_decoder
from base64 import b64decode, b64encode
from twisted.conch.ssh import keys
from hexcode import hexDecode, hexEncode

PADDING_PKCS1 = 0
PADDING_PKCS1_OAEP = 1

class RSAError(Exception):
    def __init__( self, *args ) :
        Exception.__init__( self, *args )

class RSAKey:
    def __init__(self):
        self.key = None
        
    def fromPEM_PublicKey(self, publicKeyData):
        try:
            key64 = ''
            for line in publicKeyData.splitlines():
                if line.count('BEGIN RSA PUBLIC KEY') or line.count('END RSA PUBLIC KEY'):
                    continue
                key64 += line.strip()
            pubSeq = DerSequence()
            pubSeq.decode(b64decode(key64))
            self.key = RSA.construct( (pubSeq[0], pubSeq[1]) )
        except:
            raise RSAError, 'unable to load object from data'            
        
    def toPEM_PublicKey(self):
        pemSeq = DerSequence()
        pemSeq[:] = [ self.key.key.n, self.key.key.e ]
        s = b64encode(pemSeq.encode())
        src = '-----BEGIN RSA PUBLIC KEY-----\n'
        while True:
            src += s[:64] + '\n'
            s = s[64:]
            if s == '':
                break
        src += '-----END RSA PUBLIC KEY-----'
        return src

    def fromPEM_PrivateKey(self, privKeyData, passphrase=None):
        try:
            self.key = keys.Key.fromString(privKeyData, passphrase=passphrase, type='private_openssh').keyObject
        except:
            raise RSAError, 'unable to load object from data'

    def toPEM_PrivateKey(self, passphrase=None):
        binaryKey = self.toDER_PrivateKey()
        chunks = [ binascii.b2a_base64(binaryKey[i:i+48]) for i in range(0, len(binaryKey), 48) ]
        pem = "-----BEGIN RSA PRIVATE KEY-----\n"
        pem += ''.join(chunks)
        pem += "-----END RSA PRIVATE KEY-----"
        k = keys.Key.fromString(pem, type='private_openssh')
        return k.toString('openssh', passphrase).strip()

    def fromDER_PublicKey(self, publicKeyData):
        d = der_decoder.decode(publicKeyData)
        self.key = RSA.construct((long(d[0][0]), long(d[0][1])))
        
    def toDER_PublicKey(self):
        derPK = DerSequence()
        derPK[:] = [ self.key.key.n, self.key.key.e ]
        return derPK.encode()
    
    def fromDER_PrivateKey(self, privateKeyData):
        try:
            d = der_decoder.decode(privateKeyData)
            self.key = RSA.construct((long(d[0][1]),
                                      long(d[0][2]),
                                      long(d[0][3]),
                                      long(d[0][5]),
                                      long(d[0][4]),
                                      long(d[0][8]),))
        except:
            raise RSAError, 'unable to load object from data'            
        
    def toDER_PrivateKey(self):
        derPrivKey = DerSequence()
        derPrivKey[:] = [ 0, 
                          self.key.n, 
                          self.key.e, 
                          self.key.d, 
                          self.key.q, 
                          self.key.p,
                          self.key.d % (self.key.q-1),
                          self.key.d % (self.key.p-1), 
                          self.key.u ]
        return derPrivKey.encode()
        
    def fromPKey_PublicKey(self, pkey):
        src = dump_privatekey(FILETYPE_ASN1, pkey)
        pub_der = DerSequence()
        pub_der.decode(src)
        self.key = RSA.construct((long(pub_der._seq[1]), long(pub_der._seq[2])))
    
    def construct(self, tup):
        self.key = RSA.construct(tup)
    
    def generate(self, bits=1024):
        try:
            self.key = RSA.generate(bits, os.urandom)
        except:
            raise RSAError, 'key generation failed'

    def size(self):
        return self.key.size()
    
    def check(self):
        return True
    
    def getN(self) :
        return self.key.key.n

    def getE(self) :
        return self.key.key.e

    def getD(self) :
        return self.key.key.d

    def getP(self) :
        return self.key.key.p

    def getQ(self) :
        return self.key.key.q
    
    def hasPublicKey(self) :
        return True

    def hasPrivateKey(self):
        return self.key.has_private()
    
    def getPublicKey(self) :
        return (self.key.key.n, self.key.key.e)
    
    def getPrivateKey(self):
        return  ( self.key.key.n, 
                  self.key.key.e, 
                  self.key.key.d, 
                  self.key.key.p,
                  self.key.key.q, 
                  self.key.key.d % (self.key.key.p-1),
                  self.key.key.d % (self.key.key.q-1),
                  self.key.key.u ) 
        
    def paddingSize(self, paddingMode):
        if paddingMode == 0:
            return 12
        if paddingMode == 1:
            return 42
        raise RSAError, 'unknown padding mode'
    
    def maxInputSize(self, paddingMode=1):
        return self.size() - self.paddingSize(paddingMode)
    
    def encrypt(self, data, paddingMode=1):
        try:
            return self.key.encrypt(data)
        except:
            raise RSAError, 'RSA encryption failed'
    
    def decrypt(self, data, paddingMode=1):
        try:
            return self.key.decrypt(data)
        except:
            raise RSAError, 'RSA decryption failed'
    
    def sign(self, digest, digestType):
        try:
            return self.key.sign(digest, digestType)
        except:
            raise RSAError, 'RSA sign failed'

    def verify(self, signature, digest, digestType):
        try:
            return self.key.verify(digest, (long(signature),))
        except:
            raise RSAError, 'RSA verify failed'





#!/usr/bin/python
#backup_monitor.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#This does a bunch of things.  
#1)  monitor the lists of file sent back from suppliers,
#    if there is a gap we need to try to fix it
#    * main class is _BackupMonitor,
#      it saves the lists of files in _BackupListFiles,
#      breaks down the lists of files into info on a single backup in _SupplierBackupInfo
#    * _BlockRebuilder takes care of a single broken block,
#      request what we have available, builds whatever we can
#      and stops either when we have fixed everything
#      or there is nothing more we can do
#    * _BlockRebuilder requests files through io_throttle and sends out the fixed files
#      also through io_throttle
#
#2)  if a backup is unfixable, not enough information, we delete it CleanupBackups in _BackupMonitor
#
#3)  every hour it requests a list of files from each supplier - _hourlyRequestListFiles
#
#4)  every hour it tests a file from each supplier,
#    seeing if they have the data they claim,
#    and that it is correct
#    * data is stored in _SuppliersSet and _SupplierRemoteTestResults,
#    was data good, bad, being rebuilt, or they weren't online 
#    and we got no data on the result
#    * if a supplier hasn't been seen in settings.FireInactiveSupplierIntervalHours()
#    we replace them
#
#
# Strategy for automatic backups
#
# 1) Full backup to alternating set of nodes every 2 weeks.
#
# 2) Full monthly, then incremental weekly and daily
#
# 3) One time full and then incremental monthly, weekly
#
# 4) Break alphabetical list of files into N parts and do full
#    backup on one of those and incrementals on the rest.
#    So we no longer need part of the incremental history
#      every time, and after N times older stuff can toss.
#      so every day is part full and part incremental.  Cool.
#
#  Want user to be able to specify what he wants, or at least
#  select from a few reasonable choices.
#
#
# May just do (1) to start with.
#
# This code also wakes up every day and fires off localtester, remotetester
#   on some reasonable random stuff.
#
# This manages the whole thing, so after GUI, this has the highest level control functions.
#
# Some competitors let people choose what days to backup on.  Not sure this
#   is really so great though, once we have incremental.
#
# Some can handle partial file changes.  So if a 500 MB mail file only has
# a little bit appended, only the little bit is backed up.
#
# Some competitors can turn the computer off after a backup, but
# we need it to stay on for P2P stuff.
#
# need to record if zip, tar, dump, etc
#
# If we do regular full backups often, then we might not bother scrubbing older stuff.
# Could reduce the bandwidth needed (since scrubbing could use alot.l


import os
import sys
import time
import random
import gc


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in backup_monitor.py')

from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet import threads


import lib.dhnio as dhnio
import lib.misc as misc
import lib.nameurl as nameurl
import lib.transport_control as transport_control
import lib.settings as settings
import lib.contacts as contacts
import lib.tmpfile as tmpfile
import lib.diskspace as diskspace
import lib.automat as automat
import lib.automats as automats


import backup_rebuilder
import fire_hire
import list_files_orator 
import contact_status

import identitypropagate
import backup_matrix
import backup_fs
import backup_control

_BackupMonitor = None

#------------------------------------------------------------------------------ 

def A(event=None, arg=None):
    global _BackupMonitor
    if _BackupMonitor is None:
        _BackupMonitor = BackupMonitor('backup_monitor', 'READY', 4)
    if event is not None:
        _BackupMonitor.automat(event, arg)
    return _BackupMonitor


class BackupMonitor(automat.Automat):
    timers = {'timer-1sec':   (1,       ['RESTART', 'PING']), 
              'timer-10sec':  (20,      ['PING']),
              # 'timer-10min':  (10*60,   ['READY']), 
              }
    ackCounter = 0
    pingTime = 0

    def state_changed(self, oldstate, newstate):
        automats.set_global_state('MONITOR ' + newstate)

    def A(self, event, arg):
        #---READY---
        if self.state is 'READY':
            if event == 'init' :
                backup_rebuilder.A('init')
            elif event == 'restart' :
                self.state = 'RESTART'
        #---RESTART---
        elif self.state is 'RESTART':
            if event == 'timer-1sec' and backup_rebuilder.A().state in [ 'STOPPED', 'DONE', ] :
                self.state = 'PING'
                self.doPingAllSuppliers(arg)
        #---PING---
        elif self.state is 'PING':
            if event == 'timer-10sec' or ( event == 'timer-1sec' and self.isAllSuppliersResponded(arg) ) :
                self.state = 'LIST_FILES'
                list_files_orator.A('need-files')
            elif event == 'restart' :
                self.state = 'RESTART'
        #---LIST_FILES---
        elif self.state is 'LIST_FILES':
            if event == 'restart' :
                self.state = 'RESTART'
            elif ( event == 'list_files_orator.state' and arg is 'NO_FILES' ) :
                self.state = 'READY'
            elif ( event == 'list_files_orator.state' and arg is 'SAW_FILES' ) :
                self.state = 'LIST_BACKUPS'
                self.doPrepareListBackups(arg)
        #---LIST_BACKUPS---
        elif self.state is 'LIST_BACKUPS':
            if event == 'list-backups-done' :
                self.state = 'REBUILDING'
                backup_rebuilder.A('start')
            elif event == 'restart' :
                self.state = 'RESTART'
        #---REBUILDING---
        elif self.state is 'REBUILDING':
            if event == 'restart' :
                self.state = 'RESTART'
                backup_rebuilder.SetStoppedFlag()
            elif ( event == 'backup_rebuilder.state' and arg is 'STOPPED' ) :
                self.state = 'READY'
            elif ( event == 'backup_rebuilder.state' and arg is 'DONE' ) :
                self.state = 'FIRE_HIRE'
                fire_hire.A('start')
        #---FIRE_HIRE---
        elif self.state is 'FIRE_HIRE':
            if event == 'fire-hire-finished' :
                self.state = 'READY'
                self.doCleanUpBackups(arg)
            elif event == 'restart' or event == 'hire-new-supplier' :
                self.state = 'RESTART'

    def isAllSuppliersResponded(self, arg):
        onlines = contact_status.countOnlineAmong(contacts.getSupplierIDs())
        # dhnio.Dprint(6, 'backup_monitor.isAllSuppliersResponded ackCounter=%d onlines=%d' % (self.ackCounter, onlines))
        if self.ackCounter == contacts.numSuppliers():
            return True
        if self.ackCounter >= onlines - 1:
            return True
        return False

    def doPingAllSuppliers(self, arg):
        # do not want to ping very often 
        if time.time() - self.pingTime < 60 * 3:
            self.ackCounter = contacts.numSuppliers()
            return
        self.pingTime = time.time()
        self.ackCounter = 0
        def increaseAckCounter(packet):
            self.ackCounter += 1
        dhnio.Dprint(6, 'backup_monitor.doPingAllSuppliers going to call suppliers')
        identitypropagate.suppliers(increaseAckCounter, True)

    def doPrepareListBackups(self, arg):
        # take remote and local backups and get union from it 
        allBackupIDs = set(backup_matrix.local_files().keys() + backup_matrix.remote_files().keys())
        # take only backups from data base
        allBackupIDs.intersection_update(backup_fs.ListAllBackupIDs())
        # remove running backups
        allBackupIDs.difference_update(backup_control.ListRunningBackups())
        # sort it in reverse order - newer backups should be repaired first
        allBackupIDs = misc.sorted_backup_ids(list(allBackupIDs), True)
        # add backups to the queue
        backup_rebuilder.AddBackupsToWork(allBackupIDs)
        # update suppliers, clear stats for all 
        backup_matrix.suppliers_set().UpdateSuppliers(contacts.getSupplierIDs()) 
        dhnio.Dprint(6, 'backup_monitor.doPrepareListBackups %d items' % len(allBackupIDs))
        self.automat('list-backups-done')

    def doCleanUpBackups(self, arg):
        # here we check all backups we have and remove the old one
        # user can set how many versions of that file of folder to keep 
        # other versions (older) will be removed here  
        versionsToKeep = settings.getGeneralBackupsToKeep()
        bytesUsed = backup_fs.sizebackups()/contacts.numSuppliers()
        bytesNeeded = diskspace.GetBytesFromString(settings.getCentralMegabytesNeeded(), 0) 
        dhnio.Dprint(6, 'backup_monitor.doCleanUpBackups backupsToKeep=%d used=%d needed=%d' % (versionsToKeep, bytesUsed, bytesNeeded))
        delete_count = 0
        if versionsToKeep > 0:
            for pathID, localPath, itemInfo in backup_fs.IterateIDs():
                versions = itemInfo.list_versions()
                # TODO do we need to sort the list? it comes from a set, so must be sorted may be
                while len(versions) > versionsToKeep:
                    backupID = pathID + '/' + versions.pop(0)
                    dhnio.Dprint(6, 'backup_monitor.doCleanUpBackups %d of %d backups for %s, so remove older %s' % (len(versions), versionsToKeep, localPath, backupID))
                    backup_control.DeleteBackup(backupID, saveDB=False, calculate=False)
                    delete_count += 1
        # we need also to fit used space into needed space (given from other users)
        # they trust us - do not need to take extra space from our friends
        # so remove oldest backups, but keep at least one for every folder - at least locally!
        # still our suppliers will remove our "extra" files by their "local_tester"
        if bytesNeeded <= bytesUsed:
            sizeOk = False 
            for pathID, localPath, itemInfo in backup_fs.IterateIDs():
                if sizeOk:
                    break
                versions = itemInfo.list_versions(True, False)
                if len(versions) <= 1:
                    continue
                for version in versions[1:]:
                    backupID = pathID+'/'+version
                    versionInfo = itemInfo.get_version_info(version)
                    if versionInfo[1] > 0:
                        dhnio.Dprint(6, 'backup_monitor.doCleanUpBackups over use %d of %d, so remove %s of %s' % (
                            bytesUsed, bytesNeeded, backupID, localPath))
                        backup_control.DeleteBackup(backupID, saveDB=False, calculate=False)
                        delete_count += 1
                        bytesUsed -= versionInfo[1] 
                        if bytesNeeded > bytesUsed:
                            sizeOk = True
                            break
        if delete_count > 0:
            backup_fs.Scan()
            backup_fs.Calculate()
            backup_control.Save() 
        collected = gc.collect()
        dhnio.Dprint(6, 'backup_monitor.doCleanUpBackups collected %d objects' % collected)



def Restart():
    dhnio.Dprint(4, 'backup_monitor.Restart')
    A('restart')


def shutdown():
    dhnio.Dprint(4, 'backup_monitor.shutdown')
    automat.clear_object(A().index)

        





#!/usr/bin/python
#transport_control.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
# outbox - going out to rest of world - have a packet object:
#   We keep trying methods listed for the identity we want to send to, till one gets an ACK.
#   The ACK might come back in a different way and that is fine.
#   We have a queue of packets to send for each contact, but only send one at a time each
#
# inbox - in from rest of world - input is a filename:
#   sort of the firewall point.  We try to stop garbage as much as possible.
#   We also want to keep track of recent bandwidth so others can decide which dhnpackets
#   would be easiest to get.
#
# The  transport_tcp.py  and transport_ssh.py modules are the essential ones.
# The email and q2q are not so trusted.
#
# On startup we have to start any transport receive protocols that are listed in our identity
#   on the right ports.
#
#  PREPRO - seems we are not removing temprorary files when we are done with them
#
# We can have a list of who wants what dhnpackets.
# In particular, we can have dhnblocks
# that say any packet with their number on it.



import os
import sys
import time
from time import gmtime, strftime


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in transport_control.py')

from twisted.internet.defer import Deferred,  DeferredList, succeed
from twisted.internet import task


import dhnio
import misc
import dhnnet
import settings
import contacts
import nameurl
import commands
import dhnpacket
import dhncrypto
import identitycache
# import delayeddelete
import bandwidth
import webtraffic
import tmpfile
import stun


_TransportUDPEnable = True
_TransportHTTPEnable = False
_TransportCSpaceEnable = True
_TransportEmailEnable = False
_TransportSkypeEnable = False
_TransportQ2QEnable = False
_TransportSSHEnable = False


#if dhnio.Linux():
#    _TransportQ2QEnable = False
#if dhnio.Linux():
#    _TransportSkypeEnable = False
#if dhnio.Mac():
#    _TransportSkypeEnable = False

import transport_tcp

if _TransportSSHEnable:
    import transport_ssh

if _TransportQ2QEnable:
    import transport_q2q

#if _TransportEmailEnable:
#    import transport_email

#if _TransportSkypeEnable:
#    import transport_skype

if _TransportHTTPEnable:
    import transport_http

if _TransportCSpaceEnable:
    import transport_cspace
    
if _TransportUDPEnable:
    import transport_udp

#------------------------------------------------------------------------------

#_TransportControlAutomat = None

_InitDone = False
_DoingShutdown = False
_ShutdownCount = 0
_ShutdownDeferred = None
_ProtocolsListeners = {}
_ProtocolsOptions = {}
_SupportedProtocols = set()
_ProcessSendQueueWorker = None
_StartingProtocolsSet = None
#_TimeNextSend = 0
_SendingDelay = 0.01
_LastReceiveTime = time.time()

_SendQueue = []
_InterestedParties = {}

_ContactLastAliveTime = {}
_SentFailedCountDict = {}
_SendingSpeedDict = {}
_PingContactsDict = {}
# _SendingStats = {}

_LogDebugLevel = 14
_LogFile = None
_LogSpecificProtocol = ''

_InboxPacketCallbacksList = []
_OutboxPacketCallbacksList = []
_InboxPacketStatusCallbacksList = []
_OutboxPacketStatusCallbacksList = []
_WorkItemSentCallbackList = []

_ContactAliveStateNotifierFunc = None
_MessageFunc = None
_SendFileFunc = None
_CancelFileFunc = None
_ProtocolStateNotifier = None
_PingContactFunc = None

_TrafficControlTask = None
_LongPeriodCounter = time.time()
_TrafficHistory = []

_SendingSpeed = 0
_SendingBytesCounter = 0
_SendingBytesCounterLong = 0
_SendingTimeCounter = time.time()
_SendingChunkSize = 1
_SendingSpeedLast = []

_ReceivingSpeed = 0
_ReceivingBytesCounter = 0
_ReceivingBytesCounterLong = 0
_ReceivingTimeCounter = time.time()
_ReceivingSpeedLast = []

_LiveTransfers = {}
_LastTransferID = None
_Counters = {'total_bytes': {'send': 0, 'receive': 0},
             'total_packets': {'send': 0, 'receive': 0},
             'unknown_bytes': {'send': 0, 'receive': 0},
             'unknown_packets': {'send': 0, 'receive': 0},}
_LiveTransfersByIDURL = {}

_PeersProtos = {}
_MyProtos = {}

_BlackIPs = {}

#------------------------------------------------------------------------------

# init is called on startup
# from dhninit.py or install.py
def init(init_callback=None, init_contacts=None):
    global _InitDone
    global _TransportEmailEnable
    global _TransportQ2QEnable
    global _TransportSkypeEnable
    global _StartingProtocolsSet

    dhnio.Dprint(4, 'transport_control.init')
    if _InitDone:
        dhnio.Dprint(4, 'transport_control.init already working.')
        if init_callback:
            reactor.callLater(0, init_callback)
        return

    if init_contacts is None:
        contactsDict = misc.getLocalIdentity().getContactsByProto()
    else:
        contactsDict = init_contacts

    _StartingProtocolsSet = set(contactsDict.keys())

    #--- TCP transport
    contact = contactsDict.get('tcp', None)
    if contact and settings.enableTCP():
        dhnio.Dprint(4, 'transport_control.init starting TCP transport')
        try:
            proto, host, port, filename = nameurl.UrlParse(contact)
            proto = proto.strip()
            port = misc.DigitsOnly(port)
            if port == "":
                port = settings.DefaultTCPPort()

            def _tcp_done(l, init_callback, host, port, filename):
                SetProtocolListener('tcp', l)
                SetProtocolOptions('tcp', (host, port, filename))
                SetProtocolSupport('tcp')
                ProtocolReady('tcp', init_callback)

            def _tcp_failed(x, init_callback):
                SetProtocolSupport('tcp', False)
                ProtocolReady('tcp', init_callback, False)

            d = transport_tcp.receive(int(port), receive_control_func=ReceiveTrafficControl)
            d.addCallback(_tcp_done, init_callback, host, port, filename)
            d.addErrback(_tcp_failed, init_callback)

        except:
            SetProtocolSupport('tcp', False)
            ProtocolReady('tcp', init_callback, False)
            dhnio.Dprint(1, 'transport_control.init ERROR starting transport TCP')
            dhnio.DprintException()

    else:
        _StartingProtocolsSet.discard('tcp')
        SetProtocolSupport('tcp', False)
        ProtocolReady('tcp', init_callback, False)


    #--- UDP transport
    contact = contactsDict.get('udp', None)
    if contact and settings.enableUDP() and _TransportUDPEnable:
        dhnio.Dprint(4, 'transport_control.init starting UDP transport')
        def _start_udp(x):
            if x in ['', '0.0.0.0', None]:
                SetProtocolSupport('udp', False)
                ProtocolReady('udp', init_callback, False)
                return
            try:
                host = stun.getUDPClient().externalAddress[0]
                port = str(stun.getUDPClient().externalAddress[1])
                transport_udp.init(stun.getUDPClient())
                l = transport_udp.getListener()
                SetProtocolListener('udp', l)
                SetProtocolOptions('udp', (host, port, ''))
                SetProtocolSupport('udp')
                ProtocolReady('udp', init_callback)
            except:
                SetProtocolSupport('udp', False)
                ProtocolReady('udp', init_callback, False)
                dhnio.Dprint(1, 'transport_control.init ERROR starting transport UDP')
                dhnio.DprintException()
        if stun.getUDPClient() is None:
            dhnio.Dprint(4, 'transport_control.init going to stun external IP now')
            stun.stunExternalIP(
                close_listener=False, 
                internal_port=int(settings.getUDPPort())).addBoth(
                    _start_udp)
        else:
            _start_udp('')
    else:
        SetProtocolSupport('udp', False)
        ProtocolReady('udp', init_callback, False)


    #--- SSH transport
    contact = contactsDict.get('ssh', None)
    if contact and settings.enableSSH() and _TransportSSHEnable:
        dhnio.Dprint(4, 'transport_control.init starting SSH transport')
        try:
#            contact = localIdentity.getProtoContact('ssh')
            proto, host, port, filename = nameurl.UrlParse(contact)
            proto = proto.strip()
            port = misc.DigitsOnly(port)
            if port=="":
                port = settings.DefaultSSHPort()

            transport_ssh.init()

            def _ssh_done(l, init_callback, host, port, filename):
                SetProtocolListener('ssh', l)
                SetProtocolOptions('ssh', (host, port, filename))
                SetProtocolSupport('ssh')
                ProtocolReady('ssh', init_callback)

            def _ssh_failed(x, init_callback):
                SetProtocolSupport('ssh', False)
                ProtocolReady('ssh', init_callback, False)

            d = transport_ssh.receive(int(port))
            d.addCallback(_ssh_done, init_callback, host, port, filename)
            d.addErrback(_ssh_failed, init_callback)

        except:
            SetProtocolSupport('ssh', False)
            ProtocolReady('ssh', init_callback, False)
            dhnio.Dprint(1, 'transport_control.init ERROR starting transport SSH')
            dhnio.DprintException()


    else:
        _StartingProtocolsSet.discard('ssh')
        SetProtocolSupport('ssh', False)
        ProtocolReady('ssh', init_callback, False)


    #--- Q2Q transport
    contact = contactsDict.get('q2q', None)
    if contact and _TransportQ2QEnable and settings.enableQ2Q():
        dhnio.Dprint(4, 'transport_control.init starting Q2Q transport')
        try:
            def _q2q_done(x, init_callback):
                l = transport_q2q.getListener()
                SetProtocolListener('q2q', l)
                SetProtocolOptions('q2q', (settings.getQ2Quserathost(), '', '', 'receiving'))
                SetProtocolSupport('q2q')
                ProtocolReady('q2q', init_callback)

            def _q2q_failed(x, init_callback):
                SetProtocolSupport('q2q', False)
                ProtocolReady('q2q', init_callback, False)

            d = transport_q2q.init()
            d.addCallback(_q2q_done, init_callback)
            d.addErrback(_q2q_failed, init_callback)

        except:
            SetProtocolSupport('q2q', False)
            ProtocolReady('q2q', init_callback, False)
            dhnio.Dprint(1, 'transport_control.init ERROR starting transport Q2Q')
            dhnio.DprintException()

    else:
        _StartingProtocolsSet.discard('q2q')
        SetProtocolSupport('q2q', False)
        ProtocolReady('q2q', init_callback, False)


    #--- HTTP transport
    contact = contactsDict.get('http', None)
    if contact and _TransportHTTPEnable and (settings.enableHTTP() or settings.enableHTTPServer()):
        dhnio.Dprint(4, 'transport_control.init starting HTTP transport')
        try:
            proto, host, port, filename = nameurl.UrlParse(contact)
            proto = proto.strip()
            port = misc.DigitsOnly(port)
            if port=="":
                port = settings.DefaultHTTPPort()

            transport_http.init()

            if settings.enableHTTPServer():
                transport_http.start_http_server(int(port))

            l = None
            if settings.enableHTTP() and misc.getLocalID() not in [settings.CentralID(), settings.MoneyServerID(), settings.MarketServerID()]:
                l = transport_http.receive()
            SetProtocolListener('http', l)

            SetProtocolOptions('http', (host, port, filename))
            SetProtocolSupport('http')
            ProtocolReady('http', init_callback)

        except:
            SetProtocolSupport('http', False)
            ProtocolReady('http', init_callback, False)
            dhnio.Dprint(1, 'transport_control.init ERROR starting transport HTTP')
            dhnio.DprintException()

    else:
        _StartingProtocolsSet.discard('http')
        SetProtocolSupport('http', False)
        ProtocolReady('http', init_callback, False)

    #--- CSpace transport
    contact = contactsDict.get('cspace', None)
    if contact and _TransportCSpaceEnable and settings.enableCSpace():
        dhnio.Dprint(4, 'transport_control.init starting CSPACE transport')
        try:
            def _cspace_done(x, init_callback):
                if transport_cspace.registered():
                    settings.setCSpaceKeyID(transport_cspace.keyID())
                    l = transport_cspace.getListener()
                    SetProtocolListener('cspace', l)
                    SetProtocolOptions('cspace', (settings.getCSpaceKeyID(), '', '', 'receiving'))
                    SetProtocolSupport('cspace')
                    ProtocolReady('cspace', init_callback)
                else:
                    SetProtocolSupport('cspace', False)
                    ProtocolReady('cspace', init_callback, False)
            def _cspace_failed(x, init_callback):
                SetProtocolSupport('cspace', False)
                ProtocolReady('cspace', init_callback, False)
            d = transport_cspace.init()
            d.addCallback(_cspace_done, init_callback)
            d.addErrback(_cspace_failed, init_callback)

        except:
            SetProtocolSupport('cspace', False)
            ProtocolReady('cspace', init_callback, False)
            dhnio.Dprint(1, 'transport_control.init ERROR starting transport CSpace')
            dhnio.DprintException()

    else:
        _StartingProtocolsSet.discard('cspace')
        SetProtocolSupport('cspace', False)
        ProtocolReady('cspace', init_callback, False)

    dhnio.Dprint(6, 'transport_control.init _StartingProtocolsSet=%s' % str(_StartingProtocolsSet))

    #We are done!
    _InitDone = True

#------------------------------------------------------------------------------

def AddInboxCallback(callback):
    global _InboxPacketCallbacksList
    if callback not in _InboxPacketCallbacksList:
        _InboxPacketCallbacksList.append(callback)

def AddOutboxCallback(callback):
    global _OutboxPacketCallbacksList
    if callback not in _OutboxPacketCallbacksList:
        _OutboxPacketCallbacksList.append(callback)

def AddInboxPacketStatusFunc(callback):
    global _InboxPacketStatusCallbacksList
    if callback not in _InboxPacketStatusCallbacksList:
        _InboxPacketStatusCallbacksList.append(callback)

def AddOutboxPacketStatusFunc(callback):
    global _OutboxPacketStatusCallbacksList
    if callback not in _OutboxPacketStatusCallbacksList:
        _OutboxPacketStatusCallbacksList.append(callback)
        
def AddWorkItemSentCallbackFunc(callback):
    global _WorkItemSentCallbackList
    if callback not in _WorkItemSentCallbackList:
        _WorkItemSentCallbackList.append(callback)

def SetProtocolStateNotifier(func):
    global _ProtocolStateNotifier
    _ProtocolStateNotifier = func

def SetMessageFunc(func):
    global _MessageFunc
    _MessageFunc = func

def SetContactAliveStateNotifierFunc(func):
    global _ContactAliveStateNotifierFunc
    _ContactAliveStateNotifierFunc = func
    
def GetContactAliveStateNotifierFunc():
    global _ContactAliveStateNotifierFunc
    return _ContactAliveStateNotifierFunc

def SetPingContactFunc(func):
    global _PingContactFunc
    _PingContactFunc = func

def PingContact(idurl):
    global _PingContactFunc
    if _PingContactFunc is not None:
        _PingContactFunc(idurl)

def GetOutboxFunc():
    return outboxNoAck

def RemoveFileAfterSending():
    return True

def RemoveFileAfterReceiveing():
    return True

#-------------------------------------------------------------------------------

def ListSupportedProtocols():
    global _SupportedProtocols
    return list(_SupportedProtocols)

def ProtocolIsSupported(proto):
    global _SupportedProtocols
    return proto in _SupportedProtocols

def SetProtocolSupport(proto, state=True):
    global _SupportedProtocols
    global _ProtocolStateNotifier
    if state:
        _SupportedProtocols.add(proto)
    else:
        _SupportedProtocols.discard(proto)
    if _ProtocolStateNotifier is not None:
        _ProtocolStateNotifier(proto, state)

def ProtocolListener(proto):
    global _ProtocolsListeners
    return _ProtocolsListeners.get(proto, None)

def SetProtocolListener(proto, listener):
    global _ProtocolsListeners
    _ProtocolsListeners[proto] = listener
    
def ProtocolOptions(proto):
    global _ProtocolsOptions
    return _ProtocolsOptions.get(proto, None)

def SetProtocolOptions(proto, options_tuple):
    global _ProtocolsOptions
    _ProtocolsOptions[proto] = options_tuple

def StopProtocol(proto):
    dhnio.Dprint(4, 'transport_control.StopProtocol ' + proto)
    d = None
    l = ProtocolListener(proto)
    if l is not None:
        try:
            if proto == 'tcp':
                d = l.stopListening()
            elif proto == 'ssh':
                d = l.stopListening()
            elif proto == 'q2q':
                d = l.stopListening()
            elif proto == 'http':
                if l is not None:
                    if l.active():
                        l.cancel()
            elif proto == 'cspace':
                if l is not None:
                    l.stopListening()
            elif proto == 'udp':
                if l is not None:
                    d = l.stopListening()
            else:
                dhnio.Dprint(1, 'transport_control.StopProtocol  ERROR not done YET!!!')
            del l
        except:
            dhnio.DprintException()
    SetProtocolListener(proto, None)
    SetProtocolOptions(proto, ('','','',''))
    SetProtocolSupport(proto, False)
    if d is None:
        d = Deferred()
        d.callback('')
    return d

def StartProtocol(proto, listener, host, port, filename, init_callback=None):
    dhnio.Dprint(6, 'transport_control.StartProtocol %s://%s:%s' % (proto, host, str(port)))
    SetProtocolListener(proto, listener)
    SetProtocolOptions(proto, (host, port, filename))
    SetProtocolSupport(proto)
    ProtocolReady(proto, init_callback)

#-------------------------------------------------------------------------------

def CombineIDs(CreatorID, PacketID):
    return str(CreatorID) + ":" + str(PacketID)


class InterestedParty:
    def __init__(self, CallBackFunctionOrDefer, CreatorID, PacketID):
        # dhnio.Dprint(12, "transport_control.InterestedParty making " + str(PacketID))
        # function(or Deferred)  to call when we see this packet
        self.CallBackFunction = CallBackFunctionOrDefer
        self.ComboID = CombineIDs(CreatorID, PacketID)


# cancel an interest
def RemoveInterest(CreatorID, PacketID):
    global _InterestedParties
    comboID = CombineIDs(CreatorID, PacketID)
    if _InterestedParties.has_key(comboID):
        # dhnio.Dprint(12, "transport_control.RemoveInterest comboID=" + str(comboID))
        del _InterestedParties[comboID]


# Idea is to have a list for each ComboID so that there might be more than one place called, but unique entries in that list
def RegisterInterest(CallBackFunction, CreatorID, PacketID):
    global _InterestedParties
    newparty = InterestedParty(CallBackFunction, str(CreatorID), str(PacketID))
    ExistingList = _InterestedParties.get(newparty.ComboID, "")
    if len(ExistingList) > 0:
        for oldparty in ExistingList:
            if oldparty.CallBackFunction == CallBackFunction and oldparty.ComboID == newparty.ComboID:
                return                                # already here
    if len(ExistingList) == 0:                      # if nothing there
        _InterestedParties[newparty.ComboID]=[]        #    then make it a list
    _InterestedParties[newparty.ComboID].append(newparty) # add new callback to list
    # lp=len(_InterestedParties)
    # dhnio.Dprint(12, "transport_control.RegisterInterest ComboID=%s _InterestedParties=%s " % (newparty.ComboID, str(lp)))


def FindInterestedParty(newpacket):
    global _InterestedParties
    lp = len(_InterestedParties)
    #dhnio.Dprint(12, "transport_control.FindInterestedParty  lenIP= " + str(lp))
    found = False
    ComboID = CombineIDs(newpacket.CreatorID, newpacket.PacketID)
    for party in _InterestedParties.get(ComboID, []):
        #dhnio.Dprint(12, "transport_control.FindinterestedParty Someone wants " + ComboID )
        # let him see the packet
        FuncOrDefer = party.CallBackFunction
        if isinstance(FuncOrDefer, Deferred):
            FuncOrDefer.callback(newpacket)
        else:
            FuncOrDefer(newpacket)
        found = True
    if _InterestedParties.has_key(ComboID):             # not everything has an interested party/callback
        del _InterestedParties[ComboID]                 # We called all interested parties, remove entry in dictionary
    return found


# We record interest in this packet and then send a request for it
# This is used by restore.py
def FetchAndCallBack(CallBackFunction, CreatorID, PacketID, SupplierNumber):
    RemoteID = contacts.getSupplierID(SupplierNumber)
    # dhnio.Dprint(12, "transport_control.FetchAndCallBack with RemoteID =" + RemoteID)
    RegisterInterest(CallBackFunction, CreatorID, PacketID)
    newpacket = dhnpacket.dhnpacket(commands.Retrieve(), CreatorID, CreatorID, PacketID, "", RemoteID)
    outboxNoAck(newpacket)      # send request packet - PREPRO - check that Ack really works ok
    #outbox(newpacket)          # send request packet - PREPRO - check that Ack really works ok

# deal with removing any interest in any potential data file belonging to a backup we're deleting,
# we don't want to call something trying to rebuild a backup we're deleting
def DeleteBackupInterest(BackupName):
    global _InterestedParties
    found = False
    partystoremove=[]
    for combokey in _InterestedParties.keys():
        if (combokey.find(":"+BackupName) != -1): # if the interest is for packet belonging to a backup we're dealing with
            # dhnio.Dprint(12, "transport_control.DeleteBackupInterest found for " + combokey)
            partystoremove.append(combokey)                   #   will remove party since got his callback
            found=True
    for combokey in partystoremove:                           #
        # dhnio.Dprint(12, "transport_control.DeleteBackupInterest removing " + combokey)
        del _InterestedParties[combokey]
    del partystoremove
    return found

# status can be:
#     finished     - we think we sent/received it ok
#     progress X   - sent/received X bytes so far
#     failed       - did not work for some reason


def ProtocolReady(proto, init_callback=None, state=True):
    global _SupportedProtocols
    global _ProcessSendQueueWorker
    global _StartingProtocolsSet
    #we need protocols to be ready before we start sending packets.
    #here we need to decide when we are ready.
    #just to be simple now. if we have 3 working protocols we can start
    #TODO need to improve this in future
    #Veselin changed this today
    #if first contact of the local id is ready - we are ready
    #because everybody using our first contact to send us a packets
    #Veselin changed this today again :-)
    #We want to receive ALL packets from central server when we starts
    #central will reply to our first packet using ALL our contact methods
    #so we need to be sure that ALL our transports was started
    #before we will send this first packet to central server
    #disabled transport should be counted too.

    #if we are already started - exit
    if _ProcessSendQueueWorker is not None:
        return

    #if transport_control.init() is not called yet - exit
    if _StartingProtocolsSet is None:
        return

    _StartingProtocolsSet.discard(proto)

    dhnio.Dprint(4, 'transport_control.ProtocolReady: %s with state=%s others: %s' % (str(proto), str(state), str(_StartingProtocolsSet)))

    #if all transports started - we are done
    if len(_StartingProtocolsSet) == 0:
        dhnio.Dprint(2, 'transport_control.ProtocolReady DECIDES TO START SENDING PACKETS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        #init gets this started then ProcessSendQueue keeps adding self with callLater
        StartSendQueue()

        if init_callback is not None:
            reactor.callLater(0, init_callback)


def StartSendQueue():
    global _ProcessSendQueueWorker
    dhnio.Dprint(4, 'transport_control.StartSendQueue')
    if _ProcessSendQueueWorker is not None:
        dhnio.Dprint(6, 'transport_control.StartSendQueue already started')
        return
#    _ProcessSendQueueWorker = task.LoopingCall(ProcessSendQueue)
#    _ProcessSendQueueWorker.start(0.05)
    _ProcessSendQueueWorker = reactor.callLater(0, ProcessSendQueue)


def stillActive():
    result = stillSending() or TimeSinceLastReceive < 60
    #dhnio.Dprint(14, "transport_control.stillActive " + str(result))
    return result

# PREPRO really should have a callback
# that polls to see if receiving is done also
# and does callback after
# both sending and receiving is finished.
def stillSending():
    result = SendQueueLength() > 0
    #dhnio.Dprint(14, "transport_control SendQueueLength=" + str(SendQueueLength()))
    #dhnio.Dprint(12, "transport_control.stillSending returning " + str(result))
    return result

def TimeSinceLastReceive():
    global _LastReceiveTime
    now = time.time()
    delta = now - _LastReceiveTime
    return delta

def EraseAllPeersProtosStates():
    global _PeersProtos
    _PeersProtos.clear()
    
def ErasePeerProtosStates(idurl):
    global _PeersProtos
    _PeersProtos.pop(idurl, None)
    
def EraseAllMyProtosStates():
    global _MyProtos
    _MyProtos.clear()
    
def EraseMyProtosStates(idurl):
    global _MyProtos
    _MyProtos.pop(idurl, None)

#------------------------------------------------------------------------------
# 1) The protocol modules write to temporary files and gives us that filename
# 2) We unserialize
# 3) We check that it is for us
# 4) We check that it is from one of our contacts.
# 5) We use dhnpacket.validate() to check signature and that number fields are numbers
# 6) Any other sanity checks we can do and if anything funny we toss out the packet.
# 7) Then change the filename to the dhnpacket.PackedID that it should be.
#     and call the right function(s) for this new dhnpacket
#     (dhnblock, scrubber, remotetester, customerservice, ...)
#     to dispatch it to right place(s).
#
# We have to keep track of bandwidth to/from everyone, and make a report every 24 hours
# which we send to DHN sometime in the 24 hours after that.
def inbox(filename, proto='', host=''):
    global _InboxPacketCallbacksList
    global _DoingShutdown

    if _DoingShutdown:
        dhnio.Dprint(6, "transport_control.inbox ignoring input since _DoingShutdown ")
        return None

    if filename == "" or not os.path.exists(filename):
        dhnio.Dprint(1, "transport_control.inbox  ERROR bad filename=" + filename)
        return None

    try:
        data = dhnio.ReadBinaryFile(filename)
    except:
        dhnio.Dprint(1, "transport_control.inbox ERROR reading file " + filename)
        return None

    if len(data) == 0:
        dhnio.Dprint(1, "transport_control.inbox ERROR zero byte file from %s://%s" % (proto, host))
        return None

    try:
        newpacket = dhnpacket.Unserialize(data)
    except:
        dhnio.Dprint(1, "transport_control.inbox ERROR during Unserialize data from %s://%s" % (proto, host))
        dhnio.DprintException()
        return None
    
    if newpacket is None:
        dhnio.Dprint(2, "transport_control.inbox WARNING newpacket from %s://%s is None" % (proto, host))
        return

    try:
        Command = newpacket.Command
        OwnerID = newpacket.OwnerID
        CreatorID = newpacket.CreatorID
        PacketID = newpacket.PacketID
        Date = newpacket.Date
        Payload = newpacket.Payload
        RemoteID = newpacket.RemoteID
        Signature = newpacket.Signature
        if OwnerID == misc.getLocalID() and Command == commands.Data():
            OwnerID = RemoteID
        packet_sz = len(data)
    except:
        dhnio.Dprint(1, "transport_control.inbox ERROR during Unserialize data from %s://%s" % (proto, host))
        dhnio.Dprint(1, "data length=" + str(len(data)))
        dhnio.DprintException()
        fd, filename = tmpfile.make('other', '.bad')
        os.write(fd, data)
        os.close(fd)
        return None

    dhnio.Dprint(12, "transport_control.inbox [%s] from %s, MyProtos=%s" % (Command, nameurl.GetName(OwnerID), str(_MyProtos.get(CreatorID, set()))))

    UpdateContactAliveTime(OwnerID)
    #contact_status(packet_from, 'inbox-packet', newpacket)
    
    # Might be someone who wants to look at this - remember if found someone
    foundInterest = FindInterestedParty(newpacket)

#    if newpacket.Command == commands.Ack():
#        FindAckWaitingPacket(newpacket)

    packetHandled = False
    for cb in _InboxPacketCallbacksList:
        try:
            if cb(newpacket, proto, host):
                packetHandled = True
        except:
#            # Can't get too excited about it though
            dhnio.Dprint(1, 'transport_control.inbox ERROR in callback from _InboxPacketCallbacksList: ' + str(cb))
            dhnio.DprintException()

    if not packetHandled and not foundInterest:
        dhnio.Dprint(1, "transport_control.inbox ERROR have unhandled packet [%s] %s " % (Command, PacketID))

    return newpacket


def SendAck(packettoack, response=''):
    MyID = misc.getLocalID()
    RemoteID = packettoack.OwnerID
    PacketID = packettoack.PacketID
    # dhnio.Dprint(14, "transport_control.SendAck sending to " + RemoteID)
    result = dhnpacket.dhnpacket(commands.Ack(),  MyID, MyID, PacketID, response, RemoteID)
    outboxNoAck(result)
    
def SendFail(request, response=''):
    # dhnio.Dprint(14, "transport_control.SendFail sending to " + request.OwnerID)
    result = dhnpacket.dhnpacket(commands.Fail(), misc.getLocalID(), misc.getLocalID(), request.PacketID, response, request.CreatorID)
    outboxNoAck(result)
        
    
##    dhnpacket.MakePacketInThread(outboxNoAck, commands.Ack(), MyID, MyID, PacketID, response, RemoteID)


# We can have a bunch of different "DHNPackets" that need to
# get to different IDURLs in various stages of getting out.
# The total amount of things is sort of like a window-size in
# TCP, tells you how many packets can be outstanding at once.
# Probably the idea it to keep it something that fits in RAM.
# So maybe 200 DHNPackets of 1 MB each, or something like that.
# For example, as we generate the ZIP we could do next 116 (64+52)
# DHNPackets after we got down to 84 unfinished DHNPackets.
#
# For each "DHNPackets" and an IDURL to send to,
# transport-control will work down the list of contact methods for that IRURL
# asking outbox to send by each method till we get ACK back.
# If the first is ssh we then use transport-ssh.py to send it
# (passing both DHNPackets and IDURL to that), and if email
# then we use an email-send.py, if vertex then vertex-send.py, etc.
#
# For some of these the send-*.py can get an ACK and know the
# DHNPackets got through.
#
# Point of this is to have one small file that is all someone needs
# to look at to add a new way to send DHNPackets
def outboxAck(outpacket, wide=False):
    # True = Should keep record of packet and retry till ACK comes back  -
    outbox(outpacket, True, wide)

def outboxNoAck(outpacket, wide=False):
    # False = control is not responsible for retries
    outbox(outpacket, False, wide)

def outboxAfterIdentityCaching(pagesrc, outpacket, doAck, wide=False):
    # dhnio.Dprint(12, 'transport_control.outboxAfterIdentityCaching')
    outbox(outpacket, doAck, wide)

# need to dealt with this situation
# this may happening if identity.datahaven.net is not responding
def identityCachingFailed(pagesrc, destinationID, outpacket, doAck, wide=False, count=0):
    dhnio.Dprint(6, 'transport_control.identityCachingFailed for %s NETERROR, count=%d' % (destinationID, count))
    def do_again(count):
        if count >= 3:
            return
        d = identitycache.immediatelyCaching(destinationID)
        d.addCallback(outboxAfterIdentityCaching, outpacket, doAck, wide)
        d.addErrback(identityCachingFailed, destinationID, outpacket, doAck, wide, count)
    reactor.callLater(10, do_again, count+1)

# doAck=True      if we want to wait for an Ack before counting finished
# wide=True       if we want to send packet to all contacts of Remote Identity
# retries=3       how mamy times we want to try to send the packet
def outbox(outpacket, doAck, wide=False, retries=settings.MaxRetries()):
    global _OutboxPacketCallbacksList
    global _PingContactsDict
    
    outpacket.wide = wide

    if outpacket.CreatorID != misc.getLocalID():      # if we did not make this packet
        if outpacket.Command == commands.Data():      # and it is a Data packet
            destinationID = outpacket.CreatorID       # sending someone their data back
        else:
            raise Exception("transport_control.outbox has packet we did not make that is not Data packet")
    else:
        # we made packet for someone else
        destinationID = outpacket.RemoteID

    destinationID = destinationID.strip()

    if destinationID == '':
        return

    dhnio.Dprint(12, "transport_control.outbox [%s] to %s doAck=%s wide=%s PeerProtos=%s" % (
        outpacket.Command, nameurl.GetName(outpacket.RemoteID), str(doAck), str(wide), str(_PeersProtos.get(destinationID, set()))))
    for cb in _OutboxPacketCallbacksList:
        try:
            cb(outpacket)
        except:
            dhnio.DprintException()

    #  convert ID to identity
    destinationIdentity = contacts.getContact(destinationID)
    if destinationIdentity is None:
        d = identitycache.immediatelyCaching(destinationID)
        d.addCallback(outboxAfterIdentityCaching, outpacket, doAck, wide)
        d.addErrback(identityCachingFailed, destinationID, outpacket, doAck, wide, 0)
        # Can not send without identity
        return

    fileno, filename = tmpfile.make('outbox')
    packet_data = outpacket.Serialize()
    os.write(fileno, packet_data)
    os.close(fileno)

    workitem = QueueItem(
        filename,
        len(packet_data),
        doAck,
        "presend",
        outpacket.PacketID,
        destinationID,
        outpacket.Command,
        wide,
        len(outpacket.Payload),
        retries,)
    
    SendQueueAppend(workitem)
    
    _PingContactsDict[destinationID] = time.time()
    
    del packet_data
    
    DoSend()

    # contact_status(destinationID, 'outbox-packet', outpacket)

    # dhnio.Dprint(14, "transport_control.outbox queue=%d" % SendQueueLength())


#------------------------------------------------------------------------------
# If it has stopped for a long time we might kill transfer.
# If the file is over the legal limit, we should kill transfer.
def receiveStatusReport(filename, status, proto='', host_=None, error=None, message=''):
    global _InitDone
    global _LastReceiveTime
    global _InboxPacketStatusCallbacksList
    global _MyProtos
    
    _LastReceiveTime = time.time()

    host = misc.getRealHost(host_)

    msg = message
    if error not in [None, 'None']:
        if proto != 'tcp':
            msg = '(' + str(error) + ') ' + msg

    dhnio.Dprint(8, '            [%s]     <<< %s from %s %s %s' % (proto.upper().ljust(5), status.ljust(8), host, msg, os.path.basename(filename),))

    if not _InitDone:
        return

    newpacket = inbox(filename, proto, host)

    if newpacket is not None:
        # let's remember this proto is working from this guy, 
        # because we receive from him using this proto
        if not _MyProtos.has_key(newpacket.CreatorID):
            _MyProtos[newpacket.CreatorID] = set()
        if status is 'finished':
            _MyProtos[newpacket.CreatorID].add(proto)
        
        for cb in _InboxPacketStatusCallbacksList:
            cb(newpacket, status, proto, host, error, message)

        # if settings.enableWebTraffic():
        webtraffic.inbox(newpacket, proto, host, status)

        bandwidth.INfile(filename, newpacket, proto, host, status)
        
        if status is 'finished':
            dhnnet.ConnectionDone(newpacket, proto, 'receiveStatusReport %s' % host)
        else:
            dhnnet.ConnectionFailed(newpacket, proto, 'receiveStatusReport %s' % host)

    else:
        dhnnet.ConnectionFailed(None, proto, 'receiveStatusReport %s' % host)
        try:
            fd, filename = tmpfile.make('other', '.inbox.error')
            data = dhnio.ReadBinaryFile(filename)
            os.write(fd, 'from %s:%s %s\n' % (proto, host, status))
            os.write(fd, str(data))
            os.close(fd)
        except:
            dhnio.DprintException()

    if RemoveFileAfterReceiveing():
        #delayeddelete.DeleteAfterTime(filename, 60)
        tmpfile.throw_out(filename, 'received status report')




#------------------------------------------------------------------------------
# When the transport layer has finished or failed
# it calls this so we can clean up queue and temp file
# Tricky problem is when we are doing a retransmission
# (sending same again) and get Ack from first time.
def sendStatusReport(host_, filename, status, proto='', error=None, message=''):
    global _InitDone
    global _OutboxPacketStatusCallbacksList
    global _SentFailedCountDict
    global _SendingSpeedDict
    global _PeersProtos

    host = misc.getRealHost(host_)

    msg = message
    if error not in [None, 'None']:
        if proto != 'tcp':
            msg = '(' + str(error) + ') ' + msg

    if not _InitDone:
        dhnio.Dprint(8, "            [%s] >>>     %s to %s %s %s" % (proto.upper().ljust(5), status.ljust(8), host, msg, os.path.basename(filename)))
        return

    speed = 0
    itemstoremove=[]
    for workitem in SendQueue():
        if filename == workitem.filename:
            # let's remember this proto is working to this guy
            if not _PeersProtos.has_key(workitem.remoteid):
                _PeersProtos[workitem.remoteid] = set()
            if status == 'finished':
                _PeersProtos[workitem.remoteid].add(proto)
            
            workitem.counts -= 1
            # If we don't need an Ack or already have it then we are done with this
            if workitem.doack == False or workitem.status == "acked":
                itemstoremove.append(workitem)
            else:
                # workitem.status = "sent"
                # workitem.status = 'failed' if status != 'finished' else 'sent'
                workitem.status = status.replace('finished', 'sent')

            if workitem.started is not None and status == 'finished':
                speed = float(workitem.filesize) / (time.time() - workitem.started + 0.01)
                _SendingSpeedDict[workitem.remoteid] = speed
                #dhnio.Dprint(14, "transport_control.sendStatusReport %d kb/s" % int(speed/1024))

            if not _SentFailedCountDict.has_key(workitem.remoteid):
                _SentFailedCountDict[workitem.remoteid] = 0
                
            if status is 'finished':
                _SentFailedCountDict[workitem.remoteid] = 0
                dhnnet.ConnectionDone(filename, proto, 'sendStatusReport %s' % host)
            else:
                _SentFailedCountDict[workitem.remoteid] += 1
                dhnnet.ConnectionFailed(filename, proto, 'sendStatusReport %s' % host)

            for cb in _OutboxPacketStatusCallbacksList:
                cb(workitem, proto, host, status, error, message)

            # if settings.enableWebTraffic():
            webtraffic.outbox(workitem, proto, host, status)

            bandwidth.OUTfile(filename, workitem, proto, host, status)
 
    dhnio.Dprint(8, "            [%s] >>>     %s to %s %s %s at %d kb/s" % (proto.upper().ljust(5), status.ljust(8), host, msg, os.path.basename(filename), int(speed/1024)))

    # Also removes filename
    for itemtoremove in itemstoremove:
        SendQueueRemove(itemtoremove, True, why='send status')
    del itemstoremove

#------------------------------------------------------------------------------


class QueueItem:
    def __init__(self,
            filename,
            filesize,
            doack,
            status,
            packetid,
            remoteid,
            command,
            wide,
            payloadsize,
            retries = settings.MaxRetries(),
            ):
        self.filename = filename
        self.filesize = filesize
        self.payloadsize = payloadsize
        self.doack = doack
        #"presend" - not started yet
        #"sending" - transport working on it
        #"sent" - transport says finished
        #"failed" - transport says it was failed
        #"acked" - have ack
        self.status = status
        self.packetid = packetid
        self.remoteid = remoteid
        self.command = command
        self.wide = wide
        self.time = time.time()
        self.started = None
        self.timeout = max( int(self.filesize/settings.SendingSpeedLimit() ), settings.SendTimeOut() )
        self.retries = retries
        if self.command in [ commands.Data(), ]:
            self.retries = 1
        self.counts = 0
        self.transfers = []


#------------------------------------------------------------------------------
# we want to know who are still alive

def GetContactAliveTime(idurl):
    global _ContactLastAliveTime
    return _ContactLastAliveTime.get(idurl, 0)

def TestAlive(alive_time, period=settings.DefaultAlivePacketTimeOut()*2):
    return time.time() - alive_time < period

def UpdateContactAliveTime(idurl):
    global _ContactLastAliveTime
    global _ContactAliveStateNotifierFunc
    global _SentFailedCountDict

    # we need to fire event - the contact is alive now
    # so we have to check did he was offile before
    if not TestAlive(GetContactAliveTime(idurl)):
        if _ContactAliveStateNotifierFunc is not None:
            _ContactAliveStateNotifierFunc(idurl)

    _ContactLastAliveTime[idurl] = time.time()
    _SentFailedCountDict[idurl] = 0



#------------------------------------------------------------------------------
# This is very important method
# We really need to know who is ready at the moment
# Some aspects to decide if contact is online or not:
# 1. Did not get packets from him for a long tim = offline
# 2. sendStatusReport is failed = offline
# 3. transport_cspace is tracking users - we can use it to decide
#    if user do not have cspace contact - this not fit.
#    but most users will have it!
# 4. if we check to offten - this no good, give more traffic
# 5. if we check to rarely - we have wrong info


def ContactIsAlive(idurl):
    global _SentFailedCountDict
    global _PingContactFunc
    global _PingContactsDict

    # if we haven't gotten a packet from the contact,
    # make an entry so we can have some idea
    # how long we haven't heard from him
    if GetContactAliveTime(idurl) == 0:
        ClearAliveTime(idurl)

    # if we are trying to send some packets to him
    # but it was failed - we decide that he is offline
    if _SentFailedCountDict.has_key(idurl) and _SentFailedCountDict[idurl] >= 1:
        ClearAliveTime(idurl)
        _SentFailedCountDict[idurl] = 0

    # if we did not checked this man for 15 minutes - let's do it
    # but not too often - let's check when we did it last time
    # but if this is the Central server machine - we should not do it
    if _PingContactFunc is not None:
        if misc.getLocalID() not in [settings.CentralID(), settings.MoneyServerID(), settings.MarketServerID()]:
            ping_preiod = settings.DefaultAlivePacketTimeOut() / 4
            if TestAlive(GetContactAliveTime(idurl), ping_preiod):
                if time.time()-_PingContactsDict.get(idurl, time.time()-100*ping_preiod) > ping_preiod:
                    dhnio.Dprint(8, 'transport_control.ContactIsAlive want to ping [%s]' % nameurl.GetName(idurl))
                    reactor.callLater(0, _PingContactFunc, idurl)
                    # we do update _PingContactsDict
                    # in the transport_control.outbox
                    # _PingContactsDict[idurl] = time.time()

    # check if we have some packets from contact in last hours
    return TestAlive(GetContactAliveTime(idurl))

#------------------------------------------------------------------------------


def ClearAliveTime(idurl):
    global _ContactLastAliveTime
    _ContactLastAliveTime[idurl] = time.time() - settings.DefaultAlivePacketTimeOut() * 100

def ClearAliveTimeAllContacts():
    global _ContactLastAliveTime
    for idurl in _ContactLastAliveTime.keys():
        ClearAliveTime(idurl)

def ClearAliveTimeSuppliers():
    global _ContactLastAliveTime
    for idurl in contacts.getSupplierIDs():
        if idurl:
            ClearAliveTime(idurl)

def ClearAliveTimeCustomers():
    global _ContactLastAliveTime
    for idurl in contacts.getCustomerIDs():
        ClearAliveTime(idurl)

def SendingSpeed(idurl):
    global _SendingSpeedDict
    return _SendingSpeedDict.get(idurl, 0)

#------------------------------------------------------------------------------
# Only count stuff for Active nodes - don't want to count nodes that are not answering now
# Could keep track of we have sent to but not gotten an answer from in 15 minutes or something
# Need to be careful that just loosing a packet does not qualify as being down.
# Also, if we have not sent them anything then can't complain if they don't send us anything
# So maybe keep track of last send and last receive, and "down
# Mostly want this for when doing a backup, so could just clear things at start of backup.
# Really want to know how many Data packets are waiting for sites that are Active.

def SendQueue():
    global _SendQueue
    return _SendQueue


def SendQueueActiveDataLength():
##    dhnio.Dprint(14, "transport_control.SendQueueActiveDataLength")
    datacount=0
    for workitem in SendQueue():
        if (workitem.command == commands.Data() and ContactIsAlive(workitem.remoteid)):
            datacount += 1
    return datacount


# Count all stuff on the send queue
def SendQueueLength():
    return len(SendQueue())


# Count only active stuff on the send queue
def SendQueueActiveLength():
    count = 0
    for workitem in SendQueue():
        if ContactIsAlive(workitem.remoteid):
            count += 1
    # dhnio.Dprint(14, "transport_control.SendQueueActiveLength = " + str(count))
    return count

def SendQueueAppend(workitem):
    SendQueue().append(workitem)

def SendQueueSearch(packetID):
    result = []
    for workitem in SendQueue():
        if workitem.packetid == packetID:
            result.append(workitem)
    return result

# Sometimes we can get an Ack back before transport layer reports status and is done with file.
# So we wait a few seconds to remove file.
# PREPRO Tricky problem is when we are doing a retransmission (sending same again) and get Ack from first time.
def SendQueueRemove(workitem, check_counts=False, why='removing from send queue'):
    if check_counts and workitem.counts > 0:
        return
    dhnio.Dprint(12, 'transport_control.SendQueueRemove %s to [%s] because %s' % (os.path.basename(workitem.filename), nameurl.GetName(workitem.remoteid), why))
    SendQueue().remove(workitem)
    if RemoveFileAfterSending():
        # delayeddelete.DeleteAfterTime(workitem.filename, 60)
        # if workitem.status == 'presend':
        tmpfile.throw_out(workitem.filename, why)

# If we've changed suppliers, we should remove any requests from or data to that supplier
# Only one request exactly like this, so return when found and don't worry about looping through list we modify
def RemoveSupplierRequestFromSendQueue(packetID, supplierIdentity, command):
    for workitem in SendQueue():
        if (workitem.packetid == packetID) and (workitem.remoteid == supplierIdentity) and (workitem.command == command):
            SendQueueRemove(workitem, why='remove supplier requests')
            return

# We only remove after MaxRetries
# Others are removed after sendStatusReport says sent or Ack comes back
def GetFromSendQueue():           # get the next thing that has not been sent yet
    global _SentFailedCountDict
    global _OutboxPacketStatusCallbacksList 
    for doAck in [False, True]:            # first pass we do short fast things that don't need an Ack
        for workitem in SendQueue():
            if workitem.doack == doAck:
                elapsedtime = time.time() - workitem.time
                if workitem.status == "presend":
                    workitem.status = "sending"             # leave on queue with sending status
                    workitem.time = time.time()             # need to update time when sending
                    return workitem
                elif workitem.status == "sent" and elapsedtime > workitem.timeout:
                    if workitem.retries == 0:               # we sent the file but did not get an ack yet, the time is out               
                        dhnio.Dprint(8, "transport_control.GetFromSendQueue NETERROR could not send to %s after some retries" % nameurl.GetName(workitem.remoteid))
                        SendQueueRemove(workitem, why='retries finished') # so need to remove it because retries is finished
                    else:
                        dhnio.Dprint(8, "transport_control.GetFromSendQueue doing retry to=%s elapsedtime=%s timeout=%s retry=%s" % (nameurl.GetName(workitem.remoteid), str(elapsedtime), str(workitem.timeout), str(workitem.retries)))
                        workitem.time = time.time()              # let's retry sending, need to update time when sending
                        workitem.timeout = 2 * workitem.timeout  # exponential backoff for retries
                        workitem.retries -= 1
                        return workitem
                elif workitem.status == "sending" and elapsedtime > workitem.timeout:
                    # we do sending this packet at the moment, but it seems it is too slow
                    # really don't want to retry sending it again, need to cancel this transfer at all
                    dhnio.Dprint(8, "transport_control.GetFromSendQueue NETERROR timeout sending to %s" % nameurl.GetName(workitem.remoteid))
                    for transferID in workitem.transfers:
                        CancelFile(transferID)
                    SendQueueRemove(workitem, why='timeout sending')
                elif workitem.status == 'failed':
                    dhnio.Dprint(8, "transport_control.GetFromSendQueue NETERROR failed sending to %s" % nameurl.GetName(workitem.remoteid))
                    SendQueueRemove(workitem, why='sending was failed')
                    
    return None


# We have a Queue of things to send so we can do bandwidth limit and also retry
# Command on Packet does not matter, just remoteid and packetid
def HandleSendQueueAck(packet):
    itemstoremove=[]
    for workitem in SendQueue():
        if workitem.packetid == packet.PacketID and workitem.remoteid == packet.OwnerID:
            # dhnio.Dprint(14, "transport_control.HandleSendQueueAck removing item with status " + workitem.status)
            itemstoremove.append(workitem)
    for itemtoremove in itemstoremove:
        SendQueueRemove(itemtoremove, True, why='got ack')       # Also removes filename


#  If time to, take something off the send queue
def ProcessSendQueue():
#    global _TimeNextSend
    global _ProcessSendQueueWorker
    global _SendingDelay

#    tm = time.time()
#    if tm < _TimeNextSend:
#        _ProcessSendQueueWorker = reactor.callLater(_TimeNextSend - tm, ProcessSendQueue)
#        return

    ret = DoSend()
#    if ret == 1:
#        _SendingDelay = settings.MinimumSendingDelay()
#    else:
#        if _SendingDelay < settings.MaximumSendingDelay():
#            _SendingDelay *= 2.0
    _SendingDelay = misc.LoopAttenuation(_SendingDelay, ret == 1, settings.MinimumSendingDelay(), settings.MaximumSendingDelay())
    
    # attenuation
    _ProcessSendQueueWorker = reactor.callLater(_SendingDelay, ProcessSendQueue)


def DoSend():
    global _SupportedProtocols
    global _PeersProtos

    workitem = GetFromSendQueue()
    if workitem is None:
        return 0

    if not os.path.exists(workitem.filename):
        dhnio.Dprint(1, "transport_control.DoSend ERROR should never happen that filename not exists " + workitem.filename)
        # raise Exception("transport_control.ProcessSendQueue with non existant filename=" + workitem.filename)
        return -1

    destinationIdentity = contacts.getContact(workitem.remoteid)                #  convert ID to identity
    if destinationIdentity is None:
        # dhnio.Dprint(12, "transport_control.DoSend with workitem we don't have identity for - leaving for now")
        return -1

    localIP = identitycache.GetLocalIP(workitem.remoteid)

    # here we decide which method to use
    workitem_sent = False
    if workitem.wide:
        # send to all contacts
        for contactmethod in destinationIdentity.getContacts():
            protocol, host, port, filename2 = nameurl.UrlParse(contactmethod)
            if host.strip():
                if protocol in _SupportedProtocols or protocol == 'tcp':
                    if protocol == 'tcp' and localIP:
                        host = localIP
                    SendWorkItem(workitem, protocol, host, port)
                    workitem_sent = True
    else:
        byproto = destinationIdentity.getContactsByProto()
        tcp_contact = byproto.get('tcp', None)
        cspace_contact = byproto.get('cspace', None)
        udp_contact = byproto.get('udp', None)
        working_protos = _PeersProtos.get(workitem.remoteid, set())
        # we always support tcp
        # now check if we know his local IP and he supports tcp too - than use it and send to his local IP
        if tcp_contact and localIP:
            protocol, host, port, filename2 = nameurl.UrlParse(tcp_contact)
            if host.strip():
                SendWorkItem(workitem, protocol, localIP, port)
                workitem_sent = True
        # we prefer cspace against udp, check if we did not sent the packet yet via tcp
        if not workitem_sent:
            if 'cspace' in _SupportedProtocols and cspace_contact and 'cspace' in working_protos:
                protocol, host, port, filename2 = nameurl.UrlParse(cspace_contact)
                if host.strip():
                    SendWorkItem(workitem, protocol, host, port)
                    workitem_sent = True
        # now check udp after cspace and tcp
        if not workitem_sent:
            if 'udp' in _SupportedProtocols and udp_contact and 'udp' in working_protos:
                protocol, host, port, filename2 = nameurl.UrlParse(udp_contact)
                if host.strip():
                    SendWorkItem(workitem, protocol, host, port)
                    workitem_sent = True
        # finally use the first proto we supported if we can not find the best preferable method
        if not workitem_sent: 
            for contactmethod in destinationIdentity.getContacts():
                protocol, host, port, filename2 = nameurl.UrlParse(contactmethod)
                # if method exist but empty - dont use it
                if host.strip():
                    # we always support tcp
                    if protocol in _SupportedProtocols or protocol == 'tcp':
                        # no need, it seems we already catch this above, let's try to send to his world IP 
                        # if protocol == 'tcp' and localIP:
                        #     host = localIP
                        SendWorkItem(workitem, protocol, host, port)
                        workitem_sent = True
                        break

    if not workitem_sent:
        dhnio.Dprint(6, 'transport_control.DoSend WARNING no supported protocols with ' + workitem.remoteid)
        return -1

    return 1


def SendWorkItem(workitem, protocol, host, port):
    global _TransportEmailEnable
    global _TransportQ2QEnable
    global _TransportSkypeEnable
    global _SendFileFunc
    global _WorkItemSentCallbackList

    try:
        bytes = os.path.getsize(workitem.filename)
    except:
        dhnio.DprintException()
        bytes = 0

#    OLD BANDWIDTH CONTROL
#    global _TimeNextSend
#    bandlimit = settings.BandwidthLimit() # this is kilo bytes per second
#    if bandlimit > 0:
#        bandlimit *= 1024 # switch to bytes per second
#        seconds = bytes / bandlimit # time to send the file, we send the files one by one
#        if seconds > 10:
#            seconds = 10
#    else:
#        seconds = 0
#    _TimeNextSend = time.time() + seconds

    # if we are to watch for an Ack then set that up
    if workitem.doack:
        # Retrieve is answered with our own Data packet
        if workitem.command == commands.Retrieve():
            RegisterInterest(HandleSendQueueAck, misc.getLocalID(), workitem.packetid)
        else:
            RegisterInterest(HandleSendQueueAck, workitem.remoteid, workitem.packetid)

    filename = workitem.filename

    # bandwidth.OUT(workitem.remoteid, bytes)

    # dhnio.Dprint(10, "transport_control.SendWorkItem %s [%s:%s]->%s://%s:%s (%s) retries=%d" % (os.path.basename(filename), str(workitem.command), str(workitem.packetid), protocol, host, port, nameurl.GetName(workitem.remoteid), workitem.retries))

    workitem.started = time.time()

    # if protocol == 'http':
    #     host = workitem.remoteid
    
    workitem.counts += 1
    
    _SendFileFunc(filename, protocol, host, port, workitem.remoteid, bytes, workitem.command, workitem.packetid)
    
    for cb in _WorkItemSentCallbackList:
        cb(workitem, (filename, protocol, host, port, workitem.remoteid, bytes, workitem.command, workitem.packetid))


def SendFile(filename, protocol, host, port, idurl, filesize, command, packetid):
    if not os.path.isfile(filename):
        dhnio.Dprint(8, 'transport_control.SendFile WARNING %s not found to %s://%s:%s, [%s]->[%s]' %  (os.path.basename(filename), protocol, host, port, command, nameurl.GetName(idurl)))
        sendStatusReport((host, port), filename, 'failed', protocol, None, 'file not found')
        return 
    dhnio.Dprint(8, 'transport_control.SendFile %s (%d bytes) to %s://%s:%s, [%s]->[%s]' %  (os.path.basename(filename), filesize, protocol, host, port, command, nameurl.GetName(idurl)))
    try:
        if protocol == "tcp":
            transport_tcp.send(filename, host, port, send_control_func=SendTrafficControl, description=command+'('+packetid+')')
            
        elif protocol == "udp":
            if _TransportUDPEnable:
                if command in [commands.Ack(), commands.Identity()]:
                    # those packets have higher prioriry
                    transport_udp.send(filename, host, port, True,  command+'('+packetid+')')
                else:
                    transport_udp.send(filename, host, port, False, command+'('+packetid+')')

        elif protocol == "cspace":
            if _TransportCSpaceEnable:
                transport_cspace.send(filename, host, description=command+'('+packetid+')')

#        elif protocol == "ssh":
#            idname = misc.getLocalID()
#            transport_ssh.send(host, port, idname, dhncrypto.MyPublicKey(), dhncrypto.MyPrivateKey(), filename)

#        elif protocol == "email":
#            if _TransportEmailEnable:
#                if settings.getSMTPHost() and settings.getSMTPPort():
#                    transport_email.send(filename, host)
#                else:
#                    transport_email.send_public(filename, host)

#        elif protocol == "q2q":
#            if _TransportQ2QEnable:
#                from_user = settings.getQ2Quserathost()
#                transport_q2q.send(host, filename)

#        elif protocol == "skype":
#            if _TransportSkypeEnable:
#                transport_skype.send(host, filename)

#        elif protocol == "http":
#            if _TransportHTTPEnable:
#                transport_http.send(host, filename)

        else:
            dhnio.Dprint(1, "transport_control.SendFile ERROR with protocol we don't handle yet: %s" % protocol )

    except:
        dhnio.Dprint(1, "transport_control.SendFile ERROR with exception:")
        dhnio.DprintException()


def CancelFile(transferID):
    global _LiveTransfers
    fti = _LiveTransfers.get(transferID, None)
    if not fti:
        dhnio.Dprint(2, "transport_control.CancelFile ERROR can not find transfer %s" % transferID )
        return 
    dhnio.Dprint(8, 'transport_control.CancelFile %s to %s' % (fti.filename, fti.remote_address))
    try:
        if fti.proto == "tcp":
            transport_tcp.cancel(transferID)
            
        elif fti.proto == "udp":
            if _TransportUDPEnable:
                transport_udp.cancel(transferID)

        elif fti.proto == "cspace":
            if _TransportCSpaceEnable:
                transport_cspace.cancel(transferID)

        else:
            dhnio.Dprint(2, "transport_control.CancelFile ERROR with protocol we don't handle yet: %s" % fti.proto )

    except:
        dhnio.Dprint(2, "transport_control.SendFile ERROR with exception:")
        dhnio.DprintException()


_SendFileFunc = SendFile
_CancelFileFunc = CancelFile

#------------------------------------------------------------------------------ 

# this is needed to limit outgoing bandwidth
# return number of bytes to be read now
def SendTrafficControl(prev_read_size, chunk_size): 
    global _SendingBytesCounter
    global _SendingTimeCounter
    global _SendingChunkSize
    global _SendingSpeedLast
    global _TrafficControlTask
    global _SendingSpeed
    _SendingBytesCounter += prev_read_size
    if _TrafficControlTask is None:
        # if not started yet - run traffic monitoring every second
        TrafficCounter()
    speedLimit = settings.getBandOutLimit() * 1024  # bytes per second
    if speedLimit == 0:
        _SendingChunkSize = chunk_size
        return _SendingChunkSize
    if _SendingSpeed < speedLimit:
        _SendingChunkSize = chunk_size
        return _SendingChunkSize
    dt = time.time() - _SendingTimeCounter
    if dt < 0.5 and _SendingBytesCounter > speedLimit:
        _SendingChunkSize = 1
    else:
        _SendingChunkSize = int( _SendingChunkSize / 2.0 ) or 1
    return _SendingChunkSize
     

# this is to limit incoming traffic
# return float value in seconds to wait before start reading from socket again
# accept data from multiple connections
def ReceiveTrafficControl(new_data_length):
    global _ReceivingBytesCounter
    global _ReceivingTimeCounter
    global _ReceivingSpeed
    global _TrafficControlTask
    _ReceivingBytesCounter += new_data_length
    # if not started yet - run traffic monitoring every second
    if _TrafficControlTask is None:
        TrafficCounter()
    speedLimit = settings.getBandInLimit()  * 1024  # convert to bytes per second
    if speedLimit == 0:
        return 0
    if _ReceivingSpeed <= speedLimit:
        delay = 0
    else:
        delay = ( float(_ReceivingBytesCounter) / float(_ReceivingSpeed - speedLimit) )
    dt = time.time() - _ReceivingTimeCounter
    if dt < 0.5 and _ReceivingBytesCounter > speedLimit:
        delay = ( float(_ReceivingBytesCounter) / float(speedLimit) - dt )
    if delay < 1:
        delay = 1
    if delay > 8:
        delay = 8
    return delay
 

# every second need to check the current bandwidth 
def TrafficCounter():
    global _TrafficControlTask
    global _TrafficHistory
    global _SendingSpeed
    global _ReceivingSpeed
    global _SendingBytesCounter
    global _ReceivingBytesCounter
    global _SendingBytesCounterLong
    global _ReceivingBytesCounterLong
    global _DoingShutdown
    if not _DoingShutdown:
        _TrafficControlTask = reactor.callLater(1, TrafficCounter)
    _ReceivingBytesCounterLong += _ReceivingBytesCounter
    _SendingBytesCounterLong += _SendingBytesCounter
    _TrafficHistory.append((_ReceivingBytesCounter, _SendingBytesCounter))
    if len(_TrafficHistory) > 20:
        last = _TrafficHistory.pop(0)
        _ReceivingBytesCounterLong -= last[0]
        _SendingBytesCounterLong -= last[1]
    _SendingBytesCounter = _ReceivingBytesCounter = 0
    _ReceivingSpeed = float(_ReceivingBytesCounterLong) / float(len(_TrafficHistory)) # / 1024.0
    _SendingSpeed = float(_SendingBytesCounterLong) / float(len(_TrafficHistory)) # / 1024.0
    _SendingTimeCounter = _ReceivingTimeCounter = time.time()
    # dhnio.Dprint(6, 'in: %d KB/s, out: %d KB/s, %d items' % (_ReceivingSpeed/1024.0, _SendingSpeed/1024.0, len(_TrafficHistory)))

#------------------------------------------------------------------------------ 

class FileTransferInfo():
    def __init__(self, transfer_id, proto, send_or_receive, remote_address, filename, size, callback, description):
        self.transfer_id = transfer_id
        self.proto = proto
        self.send_or_receive = send_or_receive
        self.remote_address = remote_address
        self.proto_address = '%s://%s' % (self.proto, misc.getRealHost(self.remote_address))
#        if isinstance(remote_address, tuple) and len(remote_address) == 2:  
#            self.real_address = identitycache.RemapContactAddress(self.remote_address)
#        else:
#            self.real_address = self.proto_address
        self.remote_idurl = contacts.getIDByAddress(self.proto_address)
        if self.remote_idurl is None:
            self.remote_idurl = contacts.getIDByAddress(self.remote_address)
        if self.remote_idurl is None:
            self.remote_idurl = identitycache.SearchLocalIP(self.remote_address[0])
        if self.remote_idurl is None:
            try:
                self.remote_idurl = remote_address[0]+':'+str(remote_address[1])
            except:
                self.remote_idurl = str(remote_address)        
        self.filename = filename
        self.size = size
        self.callback = callback
        self.description = description
#         dhnio.Dprint(6, 'transport_control.FileTransferInfo {%d} %s %s (%s)' % (self.transfer_id, self.send_or_receive, nameurl.GetName(self.remote_idurl), self.proto_address))
    
#    def __del__(self):
#        dhnio.Dprint(6, 'transport_control.FileTransferInfo DELETED {%d} %s %s (%s)' % (self.transfer_id, self.send_or_receive, nameurl.GetName(self.remote_idurl), self.proto_address))
        
    def get_bytes_transferred(self):
        return self.callback()

def make_transfer_ID():
    global _LastTransferID
    if _LastTransferID is None:
        _LastTransferID = int(str(int(time.time() * 100.0))[4:])
    _LastTransferID += 1
#    id = int(str(int(time.time() * 100.0))[4:])
#    if _LastTransferID == id:
#        id += 1
#    _LastTransferID = id 
    return _LastTransferID

def register_transfer(proto, send_or_receive, remote_address, callback, filename, size, description):
    global _LiveTransfers
    global _LiveTransfersByIDURL
    transfer_id = make_transfer_ID()
    fti = FileTransferInfo(transfer_id, proto, send_or_receive, remote_address, filename, size, callback, description)
    _LiveTransfers[transfer_id] = fti
    if not _LiveTransfersByIDURL.has_key(fti.remote_idurl):
        _LiveTransfersByIDURL[fti.remote_idurl] = []
    _LiveTransfersByIDURL[fti.remote_idurl].append(transfer_id)
    if send_or_receive == 'send':
        for workitem in SendQueue():
            if workitem.filename == filename:
                workitem.transfers.append(transfer_id)
    dhnio.Dprint(14, '######  transport_control.register_transfer %d [%s %s %s %s]' % (
        transfer_id, fti.send_or_receive,  fti.proto, nameurl.GetName(fti.remote_idurl), fti.proto_address))
    return transfer_id

def unregister_transfer(transfer_id):
    global _LiveTransfers
    global _LiveTransfersByIDURL
    global _Counters
    info = _LiveTransfers.pop(transfer_id, None)
    if info is None:
        dhnio.Dprint(6, '!!!!!!  transport_control.unregister_transfer WARNING %d not found, others:%d' % (transfer_id, len(_LiveTransfers)))
        return False
    b = info.get_bytes_transferred()
    _Counters['total_bytes'][info.send_or_receive] += b
    _Counters['total_packets'][info.send_or_receive] += 1
    if info.remote_idurl.startswith('http://') and info.remote_idurl.endswith('.xml'): 
        if not _Counters.has_key(info.remote_idurl):
            _Counters[info.remote_idurl] = {'send': 0, 'receive': 0}
        _Counters[info.remote_idurl][info.send_or_receive] += b
    else:
        _Counters['unknown_bytes'][info.send_or_receive] += b
        _Counters['unknown_packets'][info.send_or_receive] += 1
    if _LiveTransfersByIDURL.has_key(info.remote_idurl):
        if transfer_id in _LiveTransfersByIDURL[info.remote_idurl]:
            _LiveTransfersByIDURL[info.remote_idurl].remove(transfer_id)
            if len(_LiveTransfersByIDURL[info.remote_idurl]) == 0:
                _LiveTransfersByIDURL.pop(info.remote_idurl)
    dhnio.Dprint(14, '******  transport_control.unregister_transfer %d [%s %s %s %s]' % (
        transfer_id, info.send_or_receive, info.proto, nameurl.GetName(info.remote_idurl), info.proto_address))
    del info
    return True

def transfer_by_id(transfer_id, default=None):
    global _LiveTransfers
    return _LiveTransfers.get(transfer_id, default)

def transfers_by_idurl(idurl):
    global _LiveTransfersByIDURL
    return _LiveTransfersByIDURL.get(idurl, [])

def current_transfers():
    global _LiveTransfers
    return _LiveTransfers.values()

def current_bytes_transferred():
    global _LiveTransfers
    res = {}
    for transfer_id, info in _LiveTransfers.items():
        res[transfer_id] = info.get_bytes_transferred()
    return res

def counters():
    global _Counters
    return _Counters

#------------------------------------------------------------------------------ 

#def sending_stats():
#    global _SendingStats
#    return _SendingStats

#------------------------------------------------------------------------------ 

def black_IPs_dict():
    global _BlackIPs
    return _BlackIPs

#------------------------------------------------------------------------------ 

# Want to shutdown things cleanly so if sending we let it finish
def shutdown():
    global _ShutdownDeferred
    global _DoingShutdown
    global _InitDone
    global _InboxPacketCallbacksList
    global _OutboxPacketCallbacksList
    if not _InitDone:
        return succeed(1)
    _InboxPacketCallbacksList = []
    _OutboxPacketCallbacksList = []
    dhnio.Dprint(4, "transport_control.shutdown")
    _DoingShutdown = True
    if SendQueueActiveLength() > 0:
        dhnio.Dprint(6, "transport_control.shutdown making deferred and calling shutdown2")
        _ShutdownDeferred = Deferred()
        reactor.callLater(0, shutdown2)
        return _ShutdownDeferred
    else:
        return shutdown3()


# Here we loop waiting for sends to finish
def shutdown2():
    global _ShutdownCount
    dhnio.Dprint(4,"transport_control.shutdown2 count=" + str(_ShutdownCount))
    _ShutdownCount += 1
    if _ShutdownCount < 10 and SendQueueActiveLength() > 0:
        reactor.callLater(0.1, shutdown2)
    else:
        shutdown3()


# Here we call all the transports to tell them to shutdown and make a deferred from all their deferreds
def shutdown3():
    dhnio.Dprint(4, "transport_control.shutdown3")
    global _ProtocolsListeners
    global _ProtocolsOptions
    global _SupportedProtocols
    global _ShutdownDeferred
    global _TransportHTTPEnable
    global _InitDone
    shutlist=[]
    dhnio.Dprint(6, "transport_control.shutdown3  want to stop %d protocols " % len(_ProtocolsListeners))
    for proto, eachlistener in _ProtocolsListeners.items():
        d = StopProtocol(proto)
        shutlist.append(d)
    _ProtocolsListeners.clear()
    _ProtocolsOptions.clear()
    _SupportedProtocols.clear()
    if _TransportHTTPEnable:
        shutlist.append(transport_http.stop_http_server())
    if len(shutlist) > 0:
        result = DeferredList(shutlist)
        if _ShutdownDeferred is not None:
            result.addCallback(shutdown4)
        return result
    else:
        dhnio.Dprint(6, "transport_control.shutdown3 DeferredList seems BUGGY possible ERROR")
        _InitDone = False
        if _ShutdownDeferred is not None:
            _ShutdownDeferred.callback(1)
            return _ShutdownDeferred
        return succeed(1)


def shutdown4(result):
    global _ShutdownDeferred
    global _InitDone
    dhnio.Dprint(4, "transport_control.shutdown4")
    _InitDone = False
    if _ShutdownDeferred:
        _ShutdownDeferred.callback(result)


#-------------------------------------------------------------------------------


def main():
    dhnio.SetDebug(18)
    init()
    reactor.run()


if __name__ == '__main__':
    main()













#!/usr/bin/python
#contacts.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#  We only talk with limited sets of people and in limited ways for each type:
#   1)  datahaven.net    - central command
#   2)  supplierids      - ids for nodes who supply us with storage services (customerservice.py stuff)
#   3)  customerids      - ids for nodes we provide customerservice.py stuff for
#   4)  scrubberids      - only activated for customers who are offline more than 30 hours (or something)
#                             Rights the same as that customer number they scrub for
#   5)  correspondentids - ids for people we accept messages from
#
#  We keep around URLIDs and Identities for each.
#  We can get all the info from datahaven.net
#
#  We get a list of customers and suppliers from datahaven.net
#  We have to get an identity record for each of these contacts.
#  We have to record bandwidth between us and each contact for
#    each 24 hour period.
#  We have to report bandwidth sometime in the following 24 hour period.
#  From User/GUI or backup_monitor.py we might replace one contact at a time.
#


import os
import string
import sys


import settings
import dhnio
import misc
import contactsdb
import identitycache
import nameurl


#-------------------------------------------------------------------------------

_SuppliersChangedCallback = None
_CustomersChangedCallback = None
_ContactsChangedCallback = None
_CorrespondentsChangedCallback = None
_RequestFailsDict = {}

#------------------------------------------------------------------------------ 

#  We read from disk and if we have all the info we are set.
#  If we don't have enough, then we have to ask DHN to listcontacts and use
#  that list to get and then store all the identities for our contacts.
def init():
    dhnio.Dprint(4, "contacts.init ")
    contactsdb.load_suppliers(settings.SupplierIDsFilename())
    contactsdb.load_customers(settings.CustomerIDsFilename())
    contactsdb.load_correspondents(settings.CorrespondentIDsFilename())
    # contactsdb.add_correspondent(misc.getLocalID())

#------------------------------------------------------------------------------ 

def SetSuppliersChangedCallback(cb):
    global _SuppliersChangedCallback
    _SuppliersChangedCallback = cb

def SetCustomersChangedCallback(cb):
    global _CustomersChangedCallback
    _CustomersChangedCallback = cb

def SetContactsChangedCallback(cb):
    global _ContactsChangedCallback
    _ContactsChangedCallback = cb

def SetCorrespondentsChangedCallback(cb):
    global _CorrespondentsChangedCallback
    _CorrespondentsChangedCallback = cb

#-------------------------------------------------------------------------------

#  Sometimes we want to loop through all the IDs we make contact with
def getContactIDs():
    return contactsdb.contacts()

def IsCustomer(custid):
    return contactsdb.is_customer(custid)

def IsSupplier(supid):
    return contactsdb.is_supplier(supid)

def IsCorrespondent(idurl):
    return contactsdb.is_correspondent(idurl)

def numSuppliers():
    return contactsdb.num_suppliers()

def numCustomers():
    return contactsdb.num_customers()

def getSupplierIDs():
    return contactsdb.suppliers()

def getCustomerIDs():
    return contactsdb.customers()

def getCustomer(ID):
    if contactsdb.is_customer(ID):
        return identitycache.FromCache(ID)
    return None

def getCustomerNames():
    return map(nameurl.GetName, contactsdb.customers())

def setSupplierIDs(idslist):
    global _SuppliersChangedCallback
    global _ContactsChangedCallback
    oldsuppliers = contactsdb.suppliers()
    oldcontacts = contactsdb.contacts()
    contactsdb.set_suppliers(idslist)
    if _SuppliersChangedCallback is not None:
        _SuppliersChangedCallback(oldsuppliers, contactsdb.suppliers())
    if _ContactsChangedCallback is not None:
        _ContactsChangedCallback(oldcontacts, contactsdb.contacts())

def setCustomerIDs(idslist):
    global _CustomersChangedCallback
    global _ContactsChangedCallback
    oldcustomers = contactsdb.customers()
    oldcontacts = contactsdb.contacts()
    contactsdb.set_customers(idslist)
    if _CustomersChangedCallback is not None:
        _CustomersChangedCallback(oldcustomers, contactsdb.customers())
    if _ContactsChangedCallback is not None:
        _ContactsChangedCallback(oldcontacts, contactsdb.contacts())

def getCorrespondentIDs():
    return contactsdb.correspondents()

def getContactsAndCorrespondents():
    return contactsdb.contacts_full()

def getRemoteContacts():
    allcontacts = getContactsAndCorrespondents()
    if misc.getLocalID() in allcontacts:
        allcontacts.remove(misc.getLocalID())
    return allcontacts

def getSupplierID(N):
    return contactsdb.supplier(N)

def getSupplierN(N):
    return identitycache.FromCache(getSupplierID(N))

def getSupplier(ID):
    if contactsdb.is_supplier(ID):
        return identitycache.FromCache(ID)
    return None

def getCorrespondent(ID):
    if contactsdb.is_correspondent(ID):
        return identitycache.FromCache(ID)
    if ID in [settings.CentralID(), settings.MoneyServerID(), settings.MarketServerID()]:
        return identitycache.FromCache(ID) 
    return ''

def isKnown(idurl):
    if idurl == settings.CentralID():
        return True
    if idurl == settings.MoneyServerID():
        return True
    if idurl == settings.MarketServerID():
        return True
    if contactsdb.is_supplier(idurl):
        return True
    if contactsdb.is_customer(idurl):
        return True
    if contactsdb.is_correspondent(idurl):
        return True
    return False
    
# Suppliers should be numbered 0 to 63 with customers after that 
# not sure we can count on numbers staying
def numberForContact(idurl):
    return contactsdb.contact_index(idurl)

def numberForCustomer(idurl):
    return contactsdb.customer_index(idurl)

def numberForSupplier(idurl):
    return contactsdb.supplier_index(idurl)

def getCustomerIndex(idurl):
    return contactsdb.customer_index(idurl)

def addCustomer(idurl):
    global _CustomersChangedCallback
    global _ContactsChangedCallback
    oldcustomers = contactsdb.customers()
    oldcontacts = contactsdb.contacts()
    res = contactsdb.add_customer(idurl)
    if _CustomersChangedCallback is not None:
        _CustomersChangedCallback(oldcustomers, contactsdb.customers())
    if _ContactsChangedCallback is not None:
        _ContactsChangedCallback(oldcontacts, contactsdb.contacts())
    return res

def getCustomerID(N):
    return contactsdb.customer(N)

def getCustomerN(N):
    return identitycache.FromCache(contactsdb.customer(N))

#------------------------------------------------------------------------------ 
#--- getContact - The Main Method Here 
# Only valid contacts for dhnpackets will be signed by localidentity, suppliers, customers
# and eventually dhn central command - PREPRO
def getContact(idurl):
    global _RequestFailsDict
    if idurl is None:
        return None
    if idurl == misc.getLocalID():
        return misc.getLocalIdentity()
    if idurl == settings.CentralID():
        return identitycache.FromCache(idurl)
    if idurl == settings.MoneyServerID():
        return identitycache.FromCache(idurl)
    if idurl == settings.MarketServerID():
        return identitycache.FromCache(idurl)
    if contactsdb.is_supplier(idurl):
        return identitycache.FromCache(idurl)
    if contactsdb.is_customer(idurl):
        return identitycache.FromCache(idurl)
    if contactsdb.is_correspondent(idurl):
        return identitycache.FromCache(idurl)
    if identitycache.HasKey(idurl):
        # dhnio.Dprint(2, "contacts.getContact WARNING who is %s ?" % nameurl.GetName(idurl))
        return identitycache.FromCache(idurl)
    if misc.getLocalID() == settings.MarketServerID():
        return None
    dhnio.Dprint(2, "contacts.getContact WARNING %s not found!" % nameurl.GetName(idurl))
#    if not _RequestFailsDict.has_key(idurl):
#        _RequestFailsDict[idurl] = 0
#    _RequestFailsDict[idurl] += 1
    # TODO
    # this is not correct: 
    # need to check if other contacts is fine - if internet is turned off we can get lots fails ...  
    return None

#------------------------------------------------------------------------------ 

def getIDByAddress(address):
    idurls = identitycache.GetIDURLsByContact(address)
    if len(idurls) == 0:
        return None
    return idurls[0]
    
#------------------------------------------------------------------------------ 

def saveCustomerIDs():
    contactsdb.save_customers(settings.CustomerIDsFilename())

def saveSupplierIDs():
    contactsdb.save_suppliers(settings.SupplierIDsFilename())
    
def saveCorrespondentIDs():
    contactsdb.save_correspondents(settings.CorrespondentIDsFilename())

#-------------------------------------------------------------------------------

def addCorrespondent(idurl):
    return contactsdb.add_correspondent(idurl)
    
def removeCorrespondent(idurl):
    return contactsdb.remove_correspondent(idurl)

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    init()







#!/usr/bin/env python
#transport_ssh_server.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#

import os
import sys

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in transport_ssh_server.py')

from twisted.cred import portal
from twisted.conch import avatar
from twisted.conch.checkers import SSHPublicKeyDatabase
from twisted.conch.ssh import factory, session
from twisted.internet import protocol
from zope.interface import implements
#import tempfile


#import misc
import dhnio
import transport_control
import dhncrypto
import tmpfile


##from identity import datahaven.lib.identity as identity

#we need to keep public keys which we would like to accept on server
pub_keys_db = {}

class MyReceiveFilesSubsistem(protocol.Protocol):
    filename = ""     # string with path/filename
    fout = ""         # integer file descriptor like os.open() returns
    stop_defer = None
    total_size = 0
##    peer=""

    def __init__(self, *args, **kw):
        dhnio.Dprint(14, 'transport_ssh.MyReceiveFilesSubsistem.__init__')

    def connectionMade(self):
        transport_control.log('ssh', 'new income connection')
        dhnio.Dprint(14, 'transport_ssh.MyReceiveFilesSubsistem.connectionMade')
#        self.fout, self.filename = tempfile.mkstemp(".dhn-ssh-in")
        self.fout, self.filename = tmpfile.make("ssh-in")
        self.total_size = 0
        self.stop_defer = self.transport.session.avatar.stop_defer

    def dataReceived(self, data):
        dhnio.Dprint(14, 'transport_ssh.MyReceiveFilesSubsistem.dataReceived')
        os.write(self.fout,data)
        amount = len(data)
        self.total_size += amount
        self.transport.write(str(self.total_size))
        dhnio.Dprint(14, 'transport_ssh: total %s bytes of data received' % self.total_size)

    def connectionLost(self, reason):
        dhnio.Dprint(14, 'transport_ssh.MyReceiveFilesSubsistem.connectionLost')
        try:
            os.close(self.fout)
            transport_control.receiveStatusReport(
                self.filename,
                "finished",
                'ssh',
                self.transport.getPeer(),
                reason)
        except:
            pass
        if not self.stop_defer.called:
            self.stop_defer.callback('done')
        transport_control.log('ssh', 'income connection finished')
##        self.transport.loseConnection()


class MyAvatar(avatar.ConchUser):
    def __init__(self, username, stop_defer):
        dhnio.Dprint(14, 'transport_ssh.MyAvatar.__init__')
        avatar.ConchUser.__init__(self)
        self.username = username
        self.stop_defer = stop_defer
        self.channelLookup.update({'session':session.SSHSession})
        self.subsystemLookup.update({'sftp': MyReceiveFilesSubsistem})


class MyRealm:
    implements(portal.IRealm)

    stop_defer = None
    def __init__(self, stop_defer):
        self.stop_defer = stop_defer

    def requestAvatar(self, avatarId, mind, *interfaces):
        dhnio.Dprint(14, 'transport_ssh.MyRealm.requestAvatar')
        return interfaces[0], MyAvatar(avatarId, self.stop_defer), lambda: None

class MyPublicKeyChecker(SSHPublicKeyDatabase):
    def checkKey(self, credentials):
        dhnio.Dprint(14, 'transport_ssh.MyPublicKeyChecker.checkKey')
        global pub_keys_db
##        global not_use_publicKeyObj
##        #we do not need to accept this key
##        if credentials.blob == not_use_publicKeyObj:
##            dhnio.Dprint(1, 'transport_ssh: ERROR SECURITY VIOLATION try to login with wrong key' )
##            return False
        for name in pub_keys_db:
##            public_key = keys.getPublicKeyString(data=pub_keys_db[name])
            public_key = pub_keys_db[name]
            if credentials.blob == public_key.blob():
                dhnio.Dprint(14,'transport_ssh: user %s get access to transport_ssh server' % name)
                return True
        return False

class MyFactory(factory.SSHFactory):
##    publicKeys = {
##        'ssh-rsa': not_use_publicKeyObj.keyObject
##    }
##    privateKeys = {
####        'ssh-rsa': keys.getPrivateKeyObject(data=not_use_privateKey)
##        'ssh-rsa': keys.Key.fromString(not_use_privateKey).keyObject
##    }
##    services = {
##        'ssh-userauth': userauth.SSHUserAuthServer,
##        'ssh-connection': connection.SSHConnection
##    }

    def getPublicKeys(self):
        return { 'ssh-rsa': dhncrypto.MyPublicKeyObject() }

    def getPrivateKeys(self):
        return { 'ssh-rsa': dhncrypto.MyPrivateKeyObject() }

    def buildProtocol(self, addr):
        dhnio.Dprint(14, 'transport_ssh.MyFactory.buildProtocol ' + str(addr))
        self._transport = factory.SSHFactory.buildProtocol(self, addr)
        return self._transport

    def stopFactory(self):
        dhnio.Dprint(14, 'transport_ssh.MyFactory.stopFactory')
        try:
            self._transport.loseConnection()
        except:
            pass








from socket import socket, AF_INET, SOCK_DGRAM
from socket import error as sock_error
from ncrypt0.rand import bytes as rand_bytes
from nitro.async import AsyncOp
from nitro.tcp import tcpListen
from cspace.dht.rpc import RPCSocket
from cspace.dht.firewalltest import FirewallTestServer
from cspace.dht.client import DHTClient
from cspace.dht.node import DHTNode
from cspace.network.localip import getLocalIP
from cspace.network.router import Router

NODE_CHECK_TIME_DELAY = 60*60

class Dummy : pass

def _testFirewall( tcpListener, rpcSock, nodeAddrList, localIP,
        reactor, callback=None ) :
    def doCancel() :
        for checkOp in checkOps :
            checkOp.cancel()
        checkOps.clear()
        testServer.close()
    def onCheck( checkOp, err, payload ) :
        checkOps.remove( checkOp )
        if err >= 0 :
            fwResult,fwToken = payload
            if (fwResult >= 0) and (fwToken == token) :
                successList.append( 1 )
        if not checkOps :
            testServer.close()
            # n(success)/n(nodes) >= 1/2
            op.notify( 2*len(successList) >= len(nodeAddrList) )
    def doCheck( nodeAddr ) :
        checkOp = dhtClient.callFirewallCheck( localIP, nodeAddr,
                lambda e,p : onCheck(checkOp,e,p) )
        checkOps.add( checkOp )
    assert nodeAddrList
    successList = []
    checkOps = set()
    token = rand_bytes( 20 )
    dhtClient = DHTClient( rpcSock )
    testServer = FirewallTestServer( tcpListener, token, reactor )
    for nodeAddr in nodeAddrList :
        doCheck( nodeAddr )
    op = AsyncOp( callback, doCancel )
    return op

class NodeRunner( object ) :
    def __init__( self, nodeTable, reactor ) :
        self.nodeTable = nodeTable
        self.reactor = reactor
        self.timerOp = None
        self.firewallOp = None
        self.rpcSock = None
        self.tcpListener = None
        self.router = None
        self.dhtNode = None
        self._initTimer()

    def close( self ) :
        if self.timerOp :
            self.timerOp.cancel()
            self.timerOp = None
        if self.firewallOp :
            self.firewallOp.cancel()
            self.firewallOp = None
        if self.rpcSock :
            self.rpcSock.close()
            self.rpcSock = None
        if self.tcpListener :
            self.tcpListener.close()
            self.tcpListener = None
        if self.router is not None :
            self.router.close()
            self.router = None
        if self.dhtNode is not None :
            self.dhtNode.close()
            self.dhtNode = None

    def _initTimer( self ) :
        timeout = NODE_CHECK_TIME_DELAY
        self.timerOp = self.reactor.callLater( timeout, self._onTimer )

    def _onTimer( self ) :
        self.timerOp = None
        localIP = getLocalIP()
        if localIP is None :
            self._initTimer()
            return
        nodeAddrList = self.nodeTable.getLiveNodes()
        if not nodeAddrList :
            self._initTimer()
            return
        tcpListener = None
        udpSock = None
        for port in xrange(10001,20000) :
            addr = (localIP,port)
            try :
                tcpListener = tcpListen( addr, self.reactor, None )
            except sock_error :
                continue
            udpSock = socket( AF_INET, SOCK_DGRAM )
            try :
                udpSock.bind( addr )
            except sock_error :
                udpSock.close()
                tcpListener.close()
                udpSock = tcpListener = None
                continue
            break
        if tcpListener is None :
            self._initTimer()
            return
        self.rpcSock = RPCSocket( udpSock, self.reactor )
        self.tcpListener = tcpListener
        self.firewallOp = _testFirewall( self.tcpListener,
                self.rpcSock, nodeAddrList, localIP, self.reactor,
                self._onFirewallCheck )
    
    def _onFirewallCheck( self, result ) :
        self.firewallOp = None
        if not result :
            self.rpcSock.close()
            self.rpcSock = None
            self.tcpListener.close()
            self.tcpListener = None
            self._initTimer()
            return
        self.router = Router( self.tcpListener, self.reactor )
        self.dhtNode = DHTNode( self.rpcSock, self.reactor,
                self.nodeTable.getLiveNodes() )
        self.tcpListener = None
        self.rpcSock = None
